#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the storage usage summary for the given directories."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Base import Script
from DIRAC import gConfig, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

seSvcClassDict = {}
infoStringLength = 1
unit = None
dmScript = None


def seSvcClass(se):
  if se not in seSvcClassDict:
    from DIRAC.Resources.Storage.StorageElement import StorageElement
    try:
      if se.endswith('HIST'):
        seSvcClassDict[se] = 'Hist'
      else:
        status = StorageElement(se).getStatus()
        if status['OK']:
          status = status['Value']
          seSvcClassDict[se] = 'Tape' if status['TapeSE'] else 'Disk' if status['DiskSE'] else 'Unknown'
        else:
          seSvcClassDict[se] = 'Unknown'
    except BaseException:
      seSvcClassDict[se] = 'LFN'
  return seSvcClassDict[se]


def orderSEs(listSEs):
  """Order SEs: LFNs, then tape Ses then disk SEs The input list can be a list
  a dict, a tuple or a set."""
  orderedSEs = ['LFN'] if 'LFN' in sorted(listSEs) else []
  orderedSEs += sorted([se for se in listSEs if se not in orderedSEs and se.endswith('-HIST')])
  orderedSEs += sorted([se for se in listSEs if se not in orderedSEs and (seSvcClass(se) == 'Tape')])
  orderedSEs += sorted([se for se in listSEs if se not in orderedSEs])
  return orderedSEs


def printSEUsage(totalUsage, grandTotal, scaleFactor):
  """Nice printout of SE usage."""
  dashes = '-' * 48
  print(dashes)
  print('%s %s %s' % ('DIRAC SE'.ljust(20), ('Size (%s)' % unit).ljust(20), 'Files'.ljust(20)))
  form = '%.1f'
  orderedSEs = orderSEs(totalUsage)
  for se in orderedSEs:
    if totalUsage[se]['Size'] / scaleFactor < 1.:
      form = '%.3f'
      break
  svcClass = ''
  sumFiles = 0
  sumSize = 0.
  for se in orderedSEs:
    newSvcClass = seSvcClass(se)
    if newSvcClass != svcClass:
      if svcClass:
        print("%s %s %s" % (('Total (%s)' % svcClass).ljust(20),
                            (form % (sumSize)).ljust(20),
                            str(sumFiles).ljust(20)))
        sumFiles = 0
        sumSize = 0.
      print(dashes)
    svcClass = newSvcClass
    usageDict = totalUsage[se]
    files = usageDict['Files']
    size = usageDict['Size'] / scaleFactor
    sumFiles += files
    sumSize += size
    print("%s %s %s" % (se.ljust(20), (form % (size)).ljust(20), str(files).ljust(20)))
  if grandTotal:
    size = grandTotal['Size'] / scaleFactor
    print("%s %s %s" % ('Total (disk)'.ljust(20),
                        (form % (size)).ljust(20),
                        str(grandTotal['Files']).ljust(20)))
  print(dashes)


def printBigTable(siteList, bigTable):
  """Print out THE big table of usage."""
  siteList.sort()
  for site in ('ARCHIVE', 'LFN'):
    if site in siteList:
      siteList.remove(site)
      siteList.insert(0, site)
  just = [20, 30, 15]
  for cond in bigTable:
    just[0] = max(just[0], len(cond) + 1)
    for processingPass in sorted(bigTable[cond]):
      just[1] = max(just[1], len(processingPass) + 1)
  prStr = 'Conditions'.ljust(just[0]) + 'ProcessingPass'.ljust(just[1])
  for site in siteList:
    prStr += site.ljust(just[2])
  print(prStr)
  grandTotal = {}
  for cond in sorted(bigTable):
    print(cond.ljust(just[0]))
    for processingPass in sorted(bigTable[cond]):
      prStr = ''.ljust(just[0]) + processingPass.ljust(just[1])
      bigTableUsage = bigTable[cond][processingPass][1]
      for site in siteList:
        if site in bigTableUsage:
          grandTotal[site] = grandTotal.setdefault(site, 0) + bigTableUsage[site]
          prStr += ('%.3f' % bigTableUsage[site]).ljust(just[2])
        elif site == 'Total':
          prStr += '0'.ljust(just[2])
        else:
          prStr += '-'.ljust(just[2])
      print(prStr)
  prStr = '\n' + ''.ljust(just[0]) + 'Grand-Total'.ljust(just[1])
  for site in siteList:
    if site in grandTotal:
      prStr += ('%.3f' % grandTotal[site]).ljust(just[2])
    else:
      prStr += '-'.ljust(just[2])
  print(prStr)


def writeInfo(str):
  global infoStringLength
  import sys
  if len(str) < 1 or str[0] != '\n':
    sys.stdout.write(' ' * infoStringLength + '\r')
  sys.stdout.write(str + '\r')
  sys.stdout.flush()
  infoStringLength = len(str) + 1


def browseBK(bkQuery, ses, scaleFactor):
  # Cannot be imported at the top because CS init
  from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE

  bkPath = bkQuery.getPath()
  if not bkQuery.getConfiguration():
    print("The Configuration should be specified in the --BKQuery option: %s" % bkPath)
    return None
  conditions = bkQuery.getBKConditions()
  if not conditions:
    print('No Conditions found for this Configuration %s' % bkPath)
    return None
  requestedEventTypes = bkQuery.getEventTypeList()
  requestedFileTypes = bkQuery.getFileTypeList()
  requestedPP = bkQuery.getProcessingPass()
  requestedConditions = bkQuery.getConditions()
  bigTable = {}
  siteList = []
  for cond in conditions:
    strCond = "Browsing conditions %s" % cond
    writeInfo(strCond)
    bkQuery.setConditions(cond)
    eventTypesForPP = bkQuery.getBKProcessingPasses()
    # print eventTypesForPP
    for processingPass in [pp for pp in eventTypesForPP if eventTypesForPP[pp]]:
      if requestedEventTypes:
        eventTypes = [t for t in requestedEventTypes if t in eventTypesForPP[processingPass]]
        if not eventTypes:
          continue
      else:
        eventTypes = eventTypesForPP[processingPass]
      strPP = " - ProcessingPass %s" % processingPass
      writeInfo(strCond + strPP)
      bkQuery.setProcessingPass(processingPass)
      totalUsage = {}
      grandTotal = {}
      allProds = []
      # print eventTypes
      for eventType in eventTypes:
        strEvtType = " - EventType %s" % str(eventType)
        writeInfo(strCond + strPP + strEvtType)
        bkQuery.setEventType(eventType)
        fileTypes = bkQuery.getBKFileTypes()
        bkQuery.setFileType(fileTypes)
        # print fileTypes
        if not fileTypes or None in fileTypes:
          continue
        prods = sorted(bkQuery.getBKProductions())
        # print prods
        strProds = " - FileTypes %s - Prods %s" % (str(fileTypes), str(prods))
        writeInfo(strCond + strPP + strEvtType + strProds)
        info = bkQuery.getNumberOfLFNs()
        nbFiles = info['NumberOfLFNs']
        size = info['LFNSize']
        totalUsage['LFN']['Size'] = totalUsage.setdefault('LFN', {}).setdefault('Size', 0) + size
        totalUsage['LFN']['Files'] = totalUsage['LFN'].setdefault('Files', 0) + nbFiles
        if 'DST' not in fileTypes:
          fileTypes.append('DST')
        if 'MDST' not in fileTypes:
          fileTypes.append('MDST')
        if prods:
          allProds += prods
          for prodID in prods:
            totalUsage, grandTotal = getStorageSummary(totalUsage, grandTotal, '', fileTypes, prodID, ses)
        bkQuery.setFileType(requestedFileTypes)
        bkQuery.setEventType(requestedEventTypes)
      writeInfo('')
      if allProds:
        allProds.sort()
        print(cond, processingPass, allProds)
        printSEUsage(totalUsage, grandTotal, scaleFactor)
        processingPass = processingPass.replace('/Real Data', '')
        bigTable.setdefault(cond, {})[processingPass] = [allProds, {}]
        bigTableUsage = bigTable[cond][processingPass][1]
        for se in totalUsage:
          if se.endswith('-ARCHIVE'):
            site = 'ARCHIVE'
          elif se == 'LFN':
            site = 'LFN'
          else:
            res = getSitesForSE(se, gridName='LCG')
            if res['OK'] and len(res['Value']) > 0:
              site = res['Value'][0]
            else:
              continue
          if site not in siteList:
            siteList.append(site)
          bigTableUsage[site] = bigTableUsage.setdefault(site, 0) + (totalUsage[se]['Size'] / scaleFactor)
      bkQuery.setProcessingPass(requestedPP)
    bkQuery.setConditions(requestedConditions)
  import datetime
  print('\n', bkQuery.getPath(), str(datetime.datetime.today()).split()[0])
  printBigTable(siteList, bigTable)


def getStorageSummary(totalUsage, grandTotal, dirName, fileTypes, prodID, ses):
  from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

  if not totalUsage:
    totalUsage = {}
  if not grandTotal:
    grandTotal = {'Files': 0, 'Size': 0}
  if not isinstance(fileTypes, type([])):
    fileTypes = [fileTypes]
  for fileType in fileTypes:
    res = StorageUsageClient().getStorageSummary(dirName, fileType, prodID, ses)
    if res['OK']:
      for se in res['Value']:
        totalUsage.setdefault(se, {'Files': 0, 'Size': 0})
        totalUsage[se]['Files'] += res['Value'][se]['Files']
        totalUsage[se]['Size'] += res['Value'][se]['Size']
        if seSvcClass(se) == 'Disk' and not se.endswith('-HIST'):
          grandTotal['Files'] += res['Value'][se]['Files']
          grandTotal['Size'] += res['Value'][se]['Size']
  return totalUsage, grandTotal


def execute(unit, minimum, depth):
  from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

  # gLogger.setLevel( 'FATAL' )
  lcg = False
  full = False
  topDirectories = False
  bkBrowse = False
  users = None
  summary = False
  rank = False
  ses = []
  for switch in Script.getUnprocessedSwitches():
    if switch[0].lower() == "u" or switch[0].lower() == "unit":
      unit = switch[1]
    elif switch[0].lower() == "l" or switch[0].lower() == "lcg":
      lcg = True
    elif switch[0].lower() == "full":
      full = True
    elif switch[0] == 'TopDirectories':
      if not topDirectories:
        topDirectories = depth
    elif switch[0] == 'Depth':
      topDirectories = int(switch[1])
    elif switch[0] == "BrowseBK":
      bkBrowse = True
    elif switch[0] == "Users":
      users = switch[1].lower()
      if users == 'all':
        summary = True
        rank = True
      else:
        users = sorted(users.split(','))
    elif switch[0] == "Summary":
      summary = True
    elif switch[0] == 'Minimum':
      minimum = float(switch[1])
    elif switch[0] == 'UnknownSE':
      ses = switch[1].split(',')

  scaleDict = {'MB': 1000 * 1000.0,
               'GB': 1000 * 1000 * 1000.0,
               'TB': 1000 * 1000 * 1000 * 1000.0,
               'PB': 1000 * 1000 * 1000 * 1000 * 1000.0}
  if unit not in scaleDict:
    Script.showHelp()
  scaleFactor = scaleDict[unit]

  ses += dmScript.getOption('SEs', [])
  sites = dmScript.getOption('Sites', [])
  for site in sites:
    res = gConfig.getOptionsDict('/Resources/Sites/LCG/%s' % site)
    if not res['OK']:
      print('Site %s not known' % site)
      Script.showHelp()
    ses.extend(res['Value']['SE'].replace(' ', '').split(','))

  # Create a bkQuery looking at all files
  if bkBrowse:
    bkQuery = dmScript.getBKQuery(visible='All')
    browseBK(bkQuery, ses, scaleFactor)
    DIRAC.exit(0)

  rpc = StorageUsageClient()
  dirs = dmScript.getOption('Directory', [])
  if users:
    if users != 'all':
      dirs += ['/lhcb/user/%s/%s' % (user[0], user) for user in users]
      minimum = 0
    else:
      res = rpc.getStorageDirectories('/lhcb/user', None, None, None)
      if not res['OK']:
        print('Error getting directories in /lhcb/user:', res['Message'])
        DIRAC.exit(2)
      dirs += sorted(set(['/'.join(d.split('/')[0:5]) for d in res['Value']]))
      users = [user for user in [d.split('/')[-1] for d in dirs] if user]
  prods = None
  # print bkQuery
  fileTypes = dmScript.getOption('FileType', [])
  if not dirs:
    dirs = ['']
    bkQuery = dmScript.getBKQuery()
    if bkQuery:
      bkQuery = dmScript.getBKQuery(visible='All')
      bkFileTypes = bkQuery.getFileTypeList()
      if bkFileTypes:
        fileTypes = bkFileTypes
      print("BK query:", bkQuery)
      if fileTypes == ['RAW']:
        # For RAW data, get the list of directories...
        dirs = bkQuery.getDirs()
      else:
        prods = sorted(bkQuery.getBKProductions())
        if not prods:
          print('No productions found for bkQuery %s' % str(bkQuery))
          DIRAC.exit(0)
        # As storageSummary deals with directories and not real file types,
        #    add DST in order to cope with old naming convention
        if fileTypes and 'FULL.DST' not in fileTypes and 'DST' not in fileTypes:
          fileTypes.append('DST')
        print("Looking for %d productions:" % len(prods), prods)
    elif fileTypes and fileTypes[0]:
      print('FileTypes:', fileTypes)

  if not prods:
    prods = ['']
  if not fileTypes:
    fileTypes = ['']
  prString = "Storage usage for "
  if sites:
    prString += "Sites %s " % str(sites)
  elif ses:
    prString += "SEs %s " % str(ses)
  if prods[0] != '':
    prString += 'productions %s ' % str(prods)
  if fileTypes[0] not in ('', None):
    prString += 'file types %s ' % str(fileTypes)
  if dirs[0] != '':
    if fileTypes == ['RAW']:
      prString += 'in %d directories' % len(dirs)
    else:
      prString += 'directories %s' % str(dirs)
  if not users:
    if prString == "Storage usage for ":
      prString += 'all SEs'
    gLogger.notice(prString)
  if full or topDirectories:
    dirData = {}
    for prodID in sorted(prods):
      for fileType in fileTypes:
        for dirName in dirs:
          res = rpc.getStorageDirectoryData(dirName, fileType, prodID, ses)
          if not res['OK']:
            print('Failed to get directories', res['Message'])
            DIRAC.exit(2)
          dirData.update(res['Value'])
    if full:
      for resDir in sorted(dirData):
        gLogger.notice(resDir + " -> ", str(dirData[resDir]))
    if topDirectories:
      gLogger.notice('Depth-4 directories:')
      topDirData = {}
      for resDir, usage in dirData.items():
        topDir = '/'.join(resDir.split('/')[:topDirectories + 1]) + '/'
        topDirData.setdefault(topDir, {'Files': 0, 'Size': 0})
        topDirData[topDir]['Files'] += usage['Files']
        topDirData[topDir]['Size'] += usage['Size']
      for resDir in sorted(topDirData):
        gLogger.notice(resDir + " -> ", str(topDirData[resDir]))

  totalUsage = None
  grandTotal = None
  usersUsage = {}
  for prodID in prods:
    for fileType in fileTypes:
      for dirName in dirs:
        if users:
          user = dirName.split('/')[-1]
          userRegistry = gConfig.getOptions('/Registry/Users/%s' % user)
          if not userRegistry['OK']:
            quota = 0
          else:
            quota = gConfig.getValue('/Registry/Users/%s/Quota' % user, 0)
            if not quota:
              quota = gConfig.getValue('/Registry/DefaultStorageQuota', 0)
          quota *= scaleDict['GB'] / scaleFactor
          totalUsage, grandTotal = getStorageSummary(0, 0, dirName, fileType, prodID, ses)
          spaceUsed = grandTotal['Size'] / scaleFactor
          if summary:
            usersUsage[user] = (spaceUsed, quota)
          else:
            print("Storage usage for user %s (quota: %.1f %s)%s" % (
                user,
                quota,
                unit,
                ' <== User no longer registered' if not quota else (' <== Over quota' if spaceUsed > quota else '')
            ))
            printSEUsage(totalUsage, grandTotal, scaleFactor)
        else:
          totalUsage, grandTotal = getStorageSummary(totalUsage, grandTotal, dirName, fileType, prodID, ses)

  if lcg:
    # Cannot be imported at the top because CS init
    from DIRAC.Resources.Storage.StorageElement import StorageElement

    tapeTotalFiles = 0
    diskTotalFiles = 0
    tapeTotalSize = 0
    diskTotalSize = 0
    for se in sorted(totalUsage):
      storageElement = StorageElement(se)
      files = totalUsage[se]['Files']
      size = totalUsage[se]['Size']
      seStatus = storageElement.getStatus()
      if seStatus['OK']:
        seStatus = seStatus['Value']
        if seStatus['TapeSE']:
          tapeTotalFiles += files
          tapeTotalSize += size
        if seStatus['DiskSE']:
          diskTotalFiles += files
          diskTotalSize += size
    print('%s %s %s' % ('Storage Type'.ljust(20),
                        ('Size (%s)' % unit).ljust(20),
                        'Files'.ljust(20)))
    print('-' * 50)
    print("%s %s %s" % ('T1D*'.ljust(20),
                        ('%.1f' % (tapeTotalSize / scaleFactor)).ljust(20),
                        str(tapeTotalFiles).ljust(20)))
    print("%s %s %s" % ('T*D1'.ljust(20),
                        ('%.1f' % (diskTotalSize / scaleFactor)).ljust(20),
                        str(diskTotalFiles).ljust(20)))
    DIRAC.exit(0)

  if not users:
    printSEUsage(totalUsage, grandTotal, scaleFactor)
  elif summary:
    if rank:
      users = sorted(usersUsage, cmp=(lambda u1, u2: int((usersUsage[u2][0] - usersUsage[u1][0]) * 1000.)))
    for user in users:
      spaceUsed, quota = usersUsage[user]
      if spaceUsed > minimum:
        print("Storage usage for user %8s: %6.3f %s (quota: %4.1f %s)%s" %
              (user, spaceUsed, unit, quota, unit, ' <== User no longer registered' if not quota else (
                  ' <== Over quota' if spaceUsed > quota else '')))

  DIRAC.exit(0)


@DIRACScript()
def main():
  global unit
  global dmScript

  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  dmScript.registerNamespaceSwitches()
  dmScript.registerSiteSwitches()

  unit = 'TB'
  minimum = 0.1
  depth = 4
  Script.registerSwitch("u:", "Unit=", "   Unit to use [%s] (MB,GB,TB,PB)" % unit)
  Script.registerSwitch("l", "LCG", "  Group results by tape and disk")
  Script.registerSwitch('', "Full", "  Output the directories matching selection")
  Script.registerSwitch('', 'TopDirectories', '  Same as --Full but group by depth-%d directories' % depth)
  Script.registerSwitch('', 'Depth=', '   Specify the depth for top directories (default: %d)' % depth)
  Script.registerSwitch('', "BrowseBK", "   Loop overall paths matching the BK query")
  Script.registerSwitch('', "Users=", "   Get storage usage for a (list of) user(s)")
  Script.registerSwitch('', 'Summary', '  Only print a summary for users')
  Script.registerSwitch(
      '',
      'Minimum=',
      "   Don't print usage for users below that usage (same units, default %.1f" %
      minimum)
  Script.registerSwitch('', 'UnknownSE=', "   Force using a non-existing SE name")

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)

  execute(unit, minimum, depth)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" pytest for BookkeepingClient
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name

import pytest
from mock import MagicMock

# sut
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
bk = BookkeepingClient()

# mocks
mockRPC = MagicMock()
mockTC = MagicMock()
mockPR = MagicMock()

# test data
Step1 = ('gauss', 'Gauss', 'v1r1',
         '/some/gauss/option/files',
         'gauss-dddb_1', 'gauss-conddb_1',
         None, 1, 'Y')
Step2 = ('gauss', 'Gauss', 'v1r1',
         '/some/gauss/option/files',
         'gauss-dddb_2', 'gauss-conddb_2',
         None, 2, 'Y')
Step3 = ('gauss', 'Gauss', 'v1r1',
         '/some/gauss/option/files',
         'fromPreviousStep', 'fromPreviousStep',
         None, 3, 'Y')
Step3Corr_1 = ('gauss', 'Gauss', 'v1r1',
               '/some/gauss/option/files',
               'gauss-dddb_1', 'gauss-conddb_1',
               None, 3, 'Y')
Step3Corr_2 = ('gauss', 'Gauss', 'v1r1',
               '/some/gauss/option/files',
               'gauss-dddb_2', 'gauss-conddb_2',
               None, 3, 'Y')
Step4 = ('gauss', 'Gauss', 'v1r1',
         '/some/gauss/option/files',
         'fromPreviousStep', 'fromPreviousStep',
         None, 4, 'Y')
Step4Corr_1 = ('gauss', 'Gauss', 'v1r1',
               '/some/gauss/option/files',
               'gauss-dddb_1', 'gauss-conddb_1',
               None, 4, 'Y')
Step4Corr_2 = ('gauss', 'Gauss', 'v1r1',
               '/some/gauss/option/files',
               'gauss-dddb_2', 'gauss-conddb_2',
               None, 4, 'Y')


# Actual tests

@pytest.mark.parametrize("mockTCResponse, mockPRResponse, expected, expectedRes", [
    ({'OK': False, 'Message': 'bof'}, {}, False, None),
    ({'OK': True, 'Value': {'Status': 'Archived'}}, {}, False, None),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': False, 'Message': 'bif'}, False, None),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': True, 'Value': [1]}, True, []),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': True, 'Value': [2, 1]}, True, [2]),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': True, 'Value': [3, 2, 1]}, True, [2, 3]),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': True, 'Value': [3, 1, 2]}, True, [3]),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}}, {'OK': True, 'Value': [4, 3, 1, 2]}, True, [3, 4]),
    ({'OK': True, 'Value': {'RequestID': 1, 'Status': 'Cleaned'}}, {'OK': True, 'Value': [4, 3, 1, 2]}, False, None),
])
def test__getPreviousProductions(mocker, mockTCResponse, mockPRResponse, expected, expectedRes):
  mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.TransformationClient",
               return_value=mockTC)
  mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.ProductionRequestClient",
               return_value=mockPR)

  mockTC.getTransformation.return_value = mockTCResponse
  mockPR.getProductionList.return_value = mockPRResponse

  res = bk._getPreviousProductions(1)
  assert res['OK'] is expected
  if res['OK']:
    assert res['Value'] == expectedRes


@pytest.mark.parametrize("steps, expected", [
    ([], []),
    ([Step1], [Step1]),
    ([Step1, Step2], [Step1, Step2]),
    ([Step1, Step2, Step3], [Step1, Step2, Step3Corr_2]),
    ([Step1, Step2, Step3, Step4], [Step1, Step2, Step3Corr_2, Step4Corr_2]),
    ([Step1, Step3, Step2, Step4], [Step1, Step3Corr_1, Step2, Step4Corr_2]),
    ([Step1, Step4, Step3, Step2], [Step1, Step4Corr_1, Step3Corr_1, Step2]),
    ([Step3, Step4, Step2, Step1], [Step3, Step4, Step2, Step1]),
    ([Step3], [Step3]),
])
def test__resolveProductionSteps(steps, expected):
  res = bk._resolveProductionSteps(steps)
  assert res == expected


@pytest.mark.parametrize("mockRPCResponse, mockTCResponse, mockPRResponse, expected, expectedRes", [
    ({'OK': False, 'Message': 'bif'}, {}, {}, False, {}),
    ({'OK': True, 'Value': []}, {}, {}, False, {}),
    ({'OK': True, 'Value': [Step1]}, {}, {}, True, [Step1]),
    ({'OK': True, 'Value': [Step1, Step2]}, {}, {}, True, [Step1, Step2]),
    ({'OK': True, 'Value': [Step1, Step3]}, {}, {}, True, [Step1, Step3Corr_1]),
    ({'OK': True, 'Value': [Step1, Step3, Step4]}, {}, {}, True, [Step1, Step3Corr_1, Step4Corr_1]),
    ({'OK': True, 'Value': [Step3]},
     {'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}},
     {'OK': True, 'Value': [2, 1]},
     True,
     [Step3]),
    ({'OK': True, 'Value': [Step3, Step4]},
     {'OK': True, 'Value': {'RequestID': 1, 'Status': 'Archived'}},
     {'OK': True, 'Value': [3, 2, 1]},
     True,
     [Step3, Step4]),
])
def test_getSteps(mocker, mockRPCResponse, mockTCResponse, mockPRResponse, expected, expectedRes):
  mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.Client._getRPC",
               return_value=mockRPC)
  mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.TransformationClient",
               return_value=mockTC)
  mocker.patch("LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient.ProductionRequestClient",
               return_value=mockPR)

  mockRPC.getSteps.return_value = mockRPCResponse
  mockTC.getTransformation.return_value = mockTCResponse
  mockPR.getProductionList.return_value = mockPRResponse

  res = bk.getSteps(1)
  assert res['OK'] is expected
  if res['OK']:
    assert res['Value'] == expectedRes



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve files of a given type for a production."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... ProdID Type' % Script.scriptName,
      'Arguments:',
      '  ProdID:   Production ID (integer)',
      '  Type:     File Type (For example: ALL, DST, SIM, DIGI, RDST, MDF)']))
  Script.parseCommandLine()
  args = Script.getPositionalArgs()

  if not len(args) == 2:
    Script.showHelp(exitCode=1)

  try:
    prodID = int(args[0])
  except BaseException:
    Script.showHelp(exitCode=1)
  filetype = args[1]

  from DIRAC import gLogger, exit as DIRACexit
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  client = BookkeepingClient()
  res = client.getProductionFiles(prodID, filetype)
  if not res['OK']:
    gLogger.error('ERROR: Failed to retrieve production files', res['Message'])
    DIRACexit(1)

  if not res['Value']:
    gLogger.notice('No files found for production %s with type %s' % (prodID, filetype))
  else:
    gLogger.notice('%s %s %s %s %s' % ('FileName'.ljust(100),
                                       'Size'.ljust(10),
                                       'GUID'.ljust(40),
                                       'Replica'.ljust(8),
                                       'Visible'.ljust(8)))
    for lfn in sorted(res['Value']):
      size = res['Value'][lfn]['FileSize']
      guid = res['Value'][lfn]['GUID']
      hasReplica = res['Value'][lfn]['GotReplica']
      visible = res['Value'][lfn]['Visible']
      gLogger.notice('%s %s %s %s %s' % (lfn.ljust(100),
                                         str(size).ljust(10),
                                         guid.ljust(40),
                                         str(hasReplica).ljust(8),
                                         str(visible).ljust(8)))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It stores the replica."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.Replica import Replica
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.ReplicaParam import ReplicaParam
from DIRAC import gLogger

__RCSID__ = "$Id$"


class ReplicaReader:
  """ReplicaReader class."""
  #############################################################################

  def __init__(self):
    pass

  #############################################################################
  @staticmethod
  def readReplica(doc, filename):
    """reads the replica information."""
    gLogger.debug("Reading Replica from" + str(filename))
    replica = Replica()
    replica.setFileName(filename)  # full path

    replicaElements = doc.getElementsByTagName("Replica")

    for node in replicaElements:
      param = ReplicaParam()

      outputfile = node.getAttributeNode('File')
      if outputfile is not None:
        param.setFile(outputfile.value)
      else:
        gLogger.warn("Missing the <file> tag in replica xml file!")

      name = node.getAttributeNode('Name')
      if name is not None:
        param.setName(name.value)
      else:
        gLogger.warn("Missing the <name> tag in replica xml file!")

      location = node.getAttributeNode('Location')
      if location is not None:
        param.setLocation(location.value)
      else:
        gLogger.warn("Missing the <location> tag in replica xml file!")

      se = node.getAttributeNode('SE')
      if se is not None:
        param.setSE(se.value)
      else:
        gLogger.warn("Missing the <SE> tag in replica xml file!")

      action = node.getAttributeNode('Action')
      if action is not None:
        param.setAction(action.value)
      else:
        gLogger.warn("Missing the <Action> tag in replica xml file!")

      replica.addParam(param)
      gLogger.info("Replica Reading fhinished succesefull!!")
      return replica



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""List all BK paths matching a wildcard path ('...' is the wildcard character,
or '*' but enclose with quotes)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import Script, ProgressBar
  from DIRAC import gLogger

  Script.registerSwitch("B:", "BKQuery=", "   Bookkeeping query path")
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=True)

  bkPaths = []
  for opt, val in Script.getUnprocessedSwitches():
    if opt == 'BKQuery':
      bkPaths = val.split(',')
  if not bkPaths:
    gLogger.error('No BK path provided...')
    Script.showHelp(exitCode=1)

  from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import getProcessingPasses, BKQuery

  for i, bkPath in enumerate(bkPaths):
    if i:
      gLogger.notice('=========================')
    bkQuery = BKQuery(bkPath.replace('Real Data', 'RealData'))
    progressBar = ProgressBar(1, title="Getting processing passes for BK path %s" % bkPath)
    processingPasses = getProcessingPasses(bkQuery)
    progressBar.endLoop()
    if processingPasses:
      gLogger.notice('\n'.join([''] + [procPass.replace('Real Data', 'RealData')
                                       for procPass in sorted(processingPasses)]))
    else:
      gLogger.notice("No processing passes matching the BK path")


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set of functions used by the DMS checking scripts."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import six
from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import removeFiles, removeReplicas, \
    registerBK2FC, printDMResult


def __removeFile(lfns):
  """Use the ScriptExecutors removeFile method."""
  if isinstance(lfns, six.string_types):
    lfns = [lfns]
  removeFiles(lfns)


def __removeReplica(lfnDict):
  """Use the ScriptExecutors removeReplicas method."""
  seLFNs = {}
  for lfn in lfnDict:
    for se in lfnDict[lfn]:
      seLFNs.setdefault(se, []).append(lfn)
  for se, lfns in seLFNs.items():  # can be an iterator
    removeReplicas(lfns, [se])
  return seLFNs


def __replaceReplica(dm, seLFNs):
  """Re-replicate replicas that had just been removed because they were bad It
  uses the DataManager instance from the ConsistencyCheck."""
  if seLFNs:
    gLogger.notice("Now replicating bad replicas...")
    success = {}
    failed = {}
    for se, lfns in seLFNs.items():  # can be an iterator
      for lfn in lfns:
        res = dm.replicateAndRegister(lfn, se)
        if res['OK']:
          success.update(res['Value']['Successful'])
          failed.update(res['Value']['Failed'])
        else:
          failed[lfn] = res['Message']

    failures = 0
    errors = {}
    for lfn, reason in failed.items():  # can be an iterator
      reason = str(reason)
      errors[reason] = errors.setdefault(reason, 0) + 1
      failures += 1
    gLogger.notice("\t%d success, %d failures%s" % (len(success), failures, ':' if failures else ''))
    for reason in errors:
      gLogger.notice('\tError %s : %d files' % (reason, errors[reason]))


def _dumpErrorAndFiles(title, lfnList, maxFiles, dumpStr, fileName, fp):
  """Print out errors and a (restricted) list of LFNs dump the whole list in a
  file."""
  gLogger.error(title)
  if not gLogger.info('\n'.join(sorted(lfnList))):
    if len(lfnList) > maxFiles:
      gLogger.error('First %d files:' % maxFiles)
    gLogger.error('\n'.join(sorted(lfnList)[0:maxFiles]))
  if fileName is None:
    return None
  if fp is None:
    fp = open(fileName, 'wt')
  dumpStr1 = dumpStr.strip().replace('\n', '')
  if not dumpStr.endswith(' '):
    dumpStr += ' '
  if not dumpStr.startswith('\n'):
    dumpStr = '\n' + dumpStr
  fp.write(dumpStr)
  fp.write(dumpStr.join([''] + sorted(lfnList)))
  gLogger.notice("Whole list available with 'grep %s %s'" % (dumpStr1, fileName))
  return fp


def _getUniqueFileName(name):
  """Get a file name that doesn't exist given a prefix."""
  suffix = ''
  nb = 0
  while True:
    fileName = name + '%s.txt' % suffix
    if not os.path.exists(fileName):
      break
    nb += 1
    suffix = '-%d' % nb
  return fileName


def doCheckFC2SE(cc, bkCheck=True, fixIt=False, replace=False, maxFiles=None, fixOption='FixIt'):
  """Method actually calling for the the check using ConsistencyChecks module
  It prints out results and calls corrective actions if required."""
  cc.checkFC2SE(bkCheck)

  if maxFiles is None:
    maxFiles = 20
  fileName = _getUniqueFileName('CheckFC2SE')
  fp = None

  if cc.inSEbutNotInFC:
    gLogger.notice('>>>>')
    gLogger.notice("Some files found in SE but not in FC")
    if fixIt:
      gLogger.notice("Going to register files from BK to FC")
      regResult = {'Successful': [], 'Failed': {}}
    for lfn, seList in cc.inSEbutNotInFC.items():  # can be an iterator
      inBK = lfn in cc.existLFNsBKRepNo or lfn in cc.existLFNsBKRepYes
      gLogger.notice('\t%s (%s): %s' % (lfn, 'in BK' if inBK else 'not in BK', ','.join(seList)))
      if fixIt:
        if inBK:
          res = registerBK2FC([lfn], seList)
          if res['OK']:
            regResult['Successful'] += res['Value']['Successful']
            regResult['Failed'].update(res['Value']['Failed'])
          else:
            regResult['Failed'][lfn] = res['Message']
        else:
          regResult['Failed'][lfn] = 'Not in BK'

    if fixIt:
      # We can remove these files from the BK-No list as we just set it back
      for lfn in regResult['Successful']:
        cc.existLFNsBKRepNo.pop(lfn, None)
      printDMResult(S_OK(regResult))
    else:
      gLogger.notice("Use --FixIt to register files to the FC if they are in BK")
    gLogger.notice('<<<<')
  elif cc.seList:
    gLogger.notice("None of the files absent in FC were found at specified SEs")

  if cc.notRegisteredAtSE:
    gLogger.notice('>>>>')
    gLogger.notice("Some files found in SE but not registered in FC")
    if fixIt:
      gLogger.notice("Going to register files in FC")
      registrationProtocol = DMSHelpers().getRegistrationProtocols()
      dm = DataManager()
      replicaTuples = []
      regRepResult = {'Successful': {}, 'Failed': {}}
      for lfn, seList in cc.notRegisteredAtSE.items():  # can be an iterator
        for se in seList:
          res = StorageElement(se).getURL(lfn, protocol=registrationProtocol)
          if res['OK']:
            pfn = res['Value']['Successful'][lfn]
            replicaTuples.append((lfn, pfn, se))
      res = dm.registerReplica(replicaTuples)
      if res['OK']:
        regRepResult['Successful'].update(res['Value']['Successful'])
        regRepResult['Failed'].update(res['Value']['Failed'])
      else:
        regRepResult['Failed'][lfn] = res['Message']
      printDMResult(S_OK(regRepResult))
    else:
      gLogger.notice("Use --FixIt to register replicas in the FC")
    gLogger.notice('<<<<')

  if cc.existLFNsBKRepNo:
    gLogger.notice('>>>>')
    affectedRuns = set(str(run) for run in cc.existLFNsBKRepNo.values() if run)  # can be an iterator
    title = "%d files are in the FC (or SE) but have replica = NO in BK:\nAffected runs: %s" % \
        (len(cc.existLFNsBKRepNo),
         ','.join(sorted(affectedRuns)) if affectedRuns else 'None')
    fp = _dumpErrorAndFiles(title, cc.existLFNsBKRepNo, maxFiles, 'InFCButBKNo', fileName, fp)
    if fixIt:
      gLogger.notice("Going to fix them, setting the replica flag")
      res = cc.bkClient.addFiles(list(cc.existLFNsBKRepNo))
      if res['OK']:
        gLogger.notice("\tSuccessfully added replica flag")
      else:
        gLogger.error('Failed to set the replica flag', res['Message'])
    else:
      gLogger.notice("Use --%s to fix it (set the replica flag)" % fixOption)
    gLogger.notice('<<<<')

  elif bkCheck:
    gLogger.notice("No files in FC with replica = NO in BK -> OK!")

  if cc.existLFNsNotInBK:
    gLogger.notice('>>>>')

    title = "%d files are in the FC but are NOT in BK" % len(cc.existLFNsNotInBK)
    fp = _dumpErrorAndFiles(title, cc.existLFNsNotInBK, maxFiles, 'InFCButNotInBK ', fileName, fp)
    if fixIt:
      gLogger.notice("Going to fix them, by removing from the FC and storage")
      __removeFile(cc.existLFNsNotInBK)
    else:
      gLogger.notice("Use --%s to fix it (remove from FC and storage)" % fixOption)
    gLogger.notice('<<<<')

  else:
    gLogger.notice("No files in FC not in BK -> OK!")

  seOK = True
  if cc.existLFNsNoSE:
    gLogger.notice('>>>>')

    seOK = False
    title = "%d files are in the BK and FC but are missing at some SEs" % len(cc.existLFNsNoSE)
    fixStr = "removing them from catalogs" if not replace else "re-replicating them"
    fp = _dumpErrorAndFiles(title, cc.existLFNsNoSE, maxFiles, 'InFCButNotInSE ', fileName, fp)
    if fixIt:
      gLogger.notice("Going to fix, " + fixStr)
      removeLfns = []
      replicasToRemove = {}
      for lfn, ses in cc.existLFNsNoSE.items():  # can be an iterator
        if ses == 'All':
          removeLfns.append(lfn)
        else:
          replicasToRemove.setdefault(lfn, []).extend(ses)
      if removeLfns:
        __removeFile(removeLfns)
      if replicasToRemove:
        seLFNs = __removeReplica(replicasToRemove)
        if replace:
          __replaceReplica(cc.dataManager, seLFNs)
    else:
      if not replace:
        fixStr += "; use --Replace if you want to re-replicate them"
      gLogger.notice("Use --%s to fix it (%s)" % (fixOption, fixStr))
    gLogger.notice('<<<<')

  else:
    gLogger.notice("No missing replicas at sites -> OK!")

  if cc.existLFNsBadFiles or cc.existLFNsNotExisting:
    gLogger.notice('>>>>')

    seOK = False
    if cc.existLFNsBadFiles:
      title = "%d files have a bad checksum" % len(cc.existLFNsBadFiles)
      fp = _dumpErrorAndFiles(title, cc.existLFNsBadFiles, maxFiles, 'BadFiles', fileName, fp)
    if cc.existLFNsNotExisting:
      title = "%d files don't exist at any SE" % len(cc.existLFNsNotExisting)
      fp = _dumpErrorAndFiles(title, cc.existLFNsNotExisting, maxFiles, 'LostFiles', fileName, fp)
    toRemove = sorted(set(cc.existLFNsBadFiles) | set(cc.existLFNsNotExisting))
    if fixIt:
      gLogger.notice("Going to fix them, removing files from catalogs and storage")
      __removeFile(toRemove)
    else:
      gLogger.notice("Use --%s to fix it (remove files from SE and catalogs)" % fixOption)
    gLogger.notice('<<<<')

  if cc.existLFNsBadReplicas:
    gLogger.notice('>>>>')

    seOK = False
    title = "%d replicas have a bad checksum" % len(cc.existLFNsBadReplicas)
    badChecksum = ['%s @ %s' % (lfn, ','.join(sorted(se)))
                   for lfn, se in cc.existLFNsBadReplicas.items()]  # can be an iterator
    fp = _dumpErrorAndFiles(title, badChecksum, maxFiles, 'BadChecksum', fileName, fp)
    fixStr = "remove replicas from SE and catalogs" if not replace else "re-replicating them"
    if fixIt:
      gLogger.notice("Going to fix, " + fixStr)
      seLFNs = __removeReplica(cc.existLFNsBadReplicas)
      if replace:
        __replaceReplica(cc.dataManager, seLFNs)
    else:
      if not replace:
        fixStr += "; use --Replace if you want to re-replicate them"
      gLogger.notice("Use --%s to fix it (%s)" % (fixOption, fixStr))
    gLogger.notice('<<<<')

  if not cc.existLFNsBadFiles and not cc.existLFNsBadReplicas:
    gLogger.notice("No replicas have a bad checksum -> OK!")
  if seOK:
    gLogger.notice("All files exist and have a correct checksum -> OK!")


def doCheckFC2BK(cc, fixFC=False, fixBK=False, listAffectedRuns=False, checkFC2SE=True):
  """Method actually calling for the the check using ConsistencyChecks module
  It prints out results and calls corrective actions if required."""
  cc.checkFC2BK()

  maxFiles = 10
  suffix = ''
  nb = 0
  baseName = 'CheckFC2BK' + ('-%s' % cc.prod if cc.prod else '')
  while True:
    fileName = baseName + '%s.txt' % suffix
    if not os.path.exists(fileName):
      break
    nb += 1
    suffix = '-%d' % nb
  fp = None
  if cc.existLFNsBKRepNo:
    gLogger.notice('>>>>')

    affectedRuns = list(set(str(run) for run in cc.existLFNsBKRepNo.values()))  # can be an iterator
    gLogger.error("%d files are in the FC but have replica = NO in BK" % len(cc.existLFNsBKRepNo))
    if checkFC2SE:
      from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
      ccAux = ConsistencyChecks()
      gLogger.notice("====== Now checking %d files from FC to SE ======" % len(cc.existLFNsBKRepNo))
      ccAux.lfns = list(cc.existLFNsBKRepNo)
      doCheckFC2SE(ccAux, bkCheck=False, fixIt=fixFC, fixOption='FixFC')
      cc.existLFNsBKRepNo = set(cc.existLFNsBKRepNo)
      cc.existLFNsBKRepNo -= set(ccAux.existLFNsNoSE)
      cc.existLFNsBKRepNo -= set(ccAux.existLFNsNotExisting)
      cc.existLFNsBKRepNo -= set(ccAux.existLFNsBadFiles)
    cc.existLFNsBKRepNo = sorted(cc.existLFNsBKRepNo)
    if cc.existLFNsBKRepNo:
      gLogger.notice("====== Completed, %d files are in the FC and SE but have replica = NO in BK ======" %
                     len(cc.existLFNsBKRepNo))
      if fp is None:
        fp = open(fileName, 'wt')
      fp.write('\nInFCButBKNo '.join([''] + sorted(cc.existLFNsBKRepNo)))
      res = cc.bkClient.getFileMetadata(cc.existLFNsBKRepNo)
      if not res['OK']:
        gLogger.fatal("Unable to get file metadata", res['Message'])
        return
      if res['Value']['Failed']:
        gLogger.error("No metadata found for some files", '%d files' % len(res['Value']['Failed']))
      success = res['Value']['Successful']
      filesInvisible = set(lfn for lfn, meta in success.items()
                           if meta['VisibilityFlag'][0].upper() == 'N')  # can be an iterator
      filesVisible = set(success) - filesInvisible
      gLogger.notice('%d files are visible, %d files are invisible' %
                     (len(filesVisible), len(filesInvisible)))
      # Try and print the whole as INFO (in case --Verbose was used).
      #   If nothing printed, print a limited number of files as ERROR
      if not gLogger.info('\n'.join('%s : Visi %s' % (lfn, success.get(lfn, {}).get('VisibilityFlag', '?'))
                                    for lfn in sorted(cc.existLFNsBKRepNo))):
        if len(cc.existLFNsBKRepNo) > maxFiles:
          gLogger.notice('First %d files:' % maxFiles)
        gLogger.error('\n'.join('%s : Visi %s' % (lfn, success.get(lfn, {}).get('VisibilityFlag', '?'))
                                for lfn in sorted(cc.existLFNsBKRepNo)[0:maxFiles]))
      if listAffectedRuns:
        gLogger.notice('Affected runs: %s' % ','.join(affectedRuns))
      gLogger.notice("Full list of files:    grep InFCButBKNo %s" % fileName)
      if fixBK:
        gLogger.notice("Going to fix them, setting the replica flag")
        res = cc.bkClient.addFiles(list(success))
        if res['OK']:
          gLogger.notice("\tSuccessfully added replica flag to %d files" % len(success))
        else:
          gLogger.error('Failed to set the replica flag', res['Message'])
      elif fixFC:
        gLogger.notice("Going to fix them, by removing from the FC and storage")
        __removeFile(success)
      else:
        gLogger.notice("Use --FixBK to fix it (set the replica flag) or --FixFC (for removing from FC and storage)")
    else:
      gLogger.notice("====== Completed, no files in the FC with replica = NO in BK ======")
    gLogger.notice('<<<<')

  else:
    gLogger.notice("No files in FC with replica = NO in BK -> OK!")

  if cc.existLFNsNotInBK:
    gLogger.notice('>>>>')

    gLogger.error("%d files are in the FC but are NOT in BK:" % len(cc.existLFNsNotInBK))
    if fp is None:
      fp = open(fileName, 'wt')
    fp.write('\nInFCNotInBK '.join([''] + sorted(cc.existLFNsNotInBK)))
    if not gLogger.info('\n'.join(sorted(cc.existLFNsNotInBK))):
      if len(cc.existLFNsNotInBK) > maxFiles:
        gLogger.notice('First %d files:' % maxFiles)
      gLogger.error('\n'.join(sorted(cc.existLFNsNotInBK[0:maxFiles])))
    gLogger.notice("Full list of files:    grep InFCNotInBK %s" % fileName)
    if fixFC:
      gLogger.notice("Going to fix them, by removing from the FC and storage")
      __removeFile(cc.existLFNsNotInBK)
    else:
      gLogger.notice("Use --FixFC to fix it (remove from FC and storage)")
    gLogger.notice('<<<<')

  else:
    gLogger.notice("No files in FC not in BK -> OK!")
  if fp is not None:
    fp.close()


def doCheckBK2FC(cc, checkAll=False, fixIt=False):
  """Method actually calling for the the check using ConsistencyChecks module
  It prints out results and calls corrective actions if required."""
  cc.checkBK2FC(checkAll)
  maxPrint = 10
  chunkSize = 100
  suffix = ''
  nb = 0
  baseName = 'CheckBK2FC' + ('-%s' % cc.prod if cc.prod else '')
  while True:
    fileName = baseName + '%s.txt' % suffix
    if not os.path.exists(fileName):
      break
    nb += 1
    suffix = '-%d' % nb
  fp = None

  if checkAll:
    if cc.existLFNsBKRepNo:
      gLogger.notice('>>>>')

      nFiles = len(cc.existLFNsBKRepNo)
      comment = "%d files are in the FC but have replica = NO in BK" % nFiles
      if nFiles > maxPrint:
        comment += ' (first %d LFNs) : \n' % maxPrint
      comment += '\n'.join(cc.existLFNsBKRepNo[:maxPrint])
      gLogger.error(comment)
      if fp is None:
        fp = open(fileName, 'wt')
      fp.write('\nInFCButBKNo '.join([''] + sorted(cc.existLFNsBKRepNo)))
      gLogger.notice("To get the full list:")
      gLogger.notice("    grep InFCButBKNo %s" % fileName)
      if fixIt:
        gLogger.notice("Setting the replica flag...")
        nFiles = 0
        for lfnChunk in breakListIntoChunks(cc.existLFNsBKRepNo, chunkSize):
          res = cc.bkClient.addFiles(lfnChunk)
          if not res['OK']:
            gLogger.error("Something wrong: %s" % res['Message'])
          else:
            nFiles += len(lfnChunk)
        gLogger.notice("Successfully set replica flag to %d files" % nFiles)
      else:
        gLogger.notice("Use option --FixIt to fix it (set the replica flag)")
      gLogger.notice('<<<<')

    else:
      gLogger.notice("No LFNs exist in the FC but have replicaFlag = No in the BK -> OK!")

  if cc.absentLFNsBKRepYes:
    gLogger.notice('>>>>')

    nFiles = len(cc.absentLFNsBKRepYes)
    comment = "%d files have replicaFlag = Yes but are not in FC:" % nFiles
    if nFiles > maxPrint:
      comment += ' (first %d LFNs) : \n' % maxPrint
    comment += '\n'.join(cc.absentLFNsBKRepYes[:maxPrint])
    gLogger.error(comment)
    if fp is None:
      fp = open(fileName, 'wt')
    fp.write('\nInBKButNotInFC '.join([''] + sorted(cc.absentLFNsBKRepYes)))
    gLogger.notice("To get the full list:")
    gLogger.notice("    grep InBKButNotInFC %s" % fileName)
    if fixIt:
      gLogger.notice("Removing the replica flag...")
      nFiles = 0
      for lfnChunk in breakListIntoChunks(cc.absentLFNsBKRepYes, chunkSize):
        res = cc.bkClient.removeFiles(lfnChunk)
        if not res['OK']:
          gLogger.error("Something wrong:", res['Message'])
        else:
          nFiles += len(lfnChunk)
      gLogger.notice("Successfully removed replica flag to %d files" % nFiles)
    else:
      gLogger.notice("Use option --FixIt to fix it (remove the replica flag)")
    gLogger.notice('<<<<')

  else:
    gLogger.notice("No LFNs have replicaFlag = Yes but are not in the FC -> OK!")
  if fp:
    fp.close()


def doCheckSE(cc, seList, fixIt=False):
  cc.checkSE(seList)

  if cc.absentLFNsInFC:
    gLogger.notice('>>>>')
    gLogger.notice('%d files are not in the FC' % len(cc.absentLFNsInFC))
    if fixIt:
      __removeFile(cc.absentLFNsInFC)
    else:
      gLogger.notice("Use --FixIt to fix it (remove from catalogs")
    gLogger.notice('<<<<')

  if cc.existLFNsNoSE:
    gLogger.notice('<<<<')
    gLogger.notice('%d files are not present at %s' % (len(cc.existLFNsNoSE), ', '.join(sorted(seList))))
    gLogger.notice('\n'.join(sorted(cc.existLFNsNoSE)))
  else:
    gLogger.notice('No LFNs missing at %s' % ', '.join(sorted(seList)))



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbScript is very similar to DIRAC Script module, but consider LHCb
environment."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from DIRAC import gLogger
from DIRAC.Core.Utilities import DErrno
from DIRAC.Workflow.Modules.Script import Script

from LHCbDIRAC.Core.Utilities.RunApplication import LbRunError


class LHCbScript(Script):
  """A simple extension to the DIRAC script module."""

  def __init__(self):
    """c'tor."""
    self.log = gLogger.getSubLogger('LHCbScript')
    super(LHCbScript, self).__init__(self.log)

    self.systemConfig = 'ANY'
    self.environment = {}

  def _resolveInputVariables(self):
    """By convention the workflow parameters are resolved here."""

    super(LHCbScript, self)._resolveInputVariables()
    super(LHCbScript, self)._resolveInputStep()

    self.systemConfig = self.step_commons.get('SystemConfig', self.systemConfig)

  def _executeCommand(self):
    """Executes the self.command (uses systemCall) with binary tag (CMTCONFIG)
    requested (if not 'ANY')"""

    if self.systemConfig != 'ANY':
      self.environment = os.environ
      self.environment['CMTCONFIG'] = self.systemConfig

    super(LHCbScript, self)._executeCommand()

  def _exitWithError(self, status):
    """Extended here for treating case of lb-run error codes (and executable
    name)."""
    # this is an lb-run specific error
    if status & 0x40 and not status & 0x80:
      self.log.error("Exit status is an lb-run specific error", '(%s)' % status)
      raise LbRunError("Problem setting the environment: lb-run exited with status %d" % status,
                       DErrno.EWMSRESC)
    else:
      super(LHCbScript, self)._exitWithError(status)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""unit tests for Configuration Helpers."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import mock
import os
import pytest

from DIRAC import gLogger

from DIRAC.Core.Security.Locations import getCAsLocation
import LHCbDIRAC.ConfigurationSystem.Client.Helpers.Resources as moduleTested

gLogger.setLevel('DEBUG')


@pytest.mark.parametrize("binaryTagsSet, expected", [
    (set(), None),
    ({frozenset({'x86_64-slc6-gcc48-opt'})}, 'x86_64-slc6'),
    ({frozenset({'x86_64-slc6-gcc62-opt'})}, 'x86_64_v2-slc6'),
    ({frozenset({'x86_64+avx2-slc6-gcc62-opt'})}, 'x86_64_v3-slc6'),
    ({frozenset({'x86_64+avx2+fma-centos7-gcc7-opt'})}, 'x86_64_v3-centos7'),
    ({frozenset({'x86_64-slc6-gcc48-opt'}),
      frozenset({'x86_64-slc6-gcc62-opt'})}, 'x86_64_v2-slc6'),
    ({frozenset({'x86_64-centos7-gcc49-opt'}),
      frozenset({'x86_64-slc6-gcc62-opt'})}, 'x86_64_v2-centos7'),
    ({frozenset({'x86_64-centos7-gcc49-opt'}),
      frozenset({'x86_64-slc6-gcc62-opt'}),
      frozenset({'x86_64+avx2-centos7-gcc62-opt'})}, 'x86_64_v3-centos7'),
    ({frozenset({'x86_64-centos7-gcc49-opt'}),
      frozenset({'x86_64+avx-slc6-gcc62-opt'}),
      frozenset({'x86_64+avx2-centos7-gcc62-opt'})}, 'x86_64_v3-centos7'),
    ({frozenset({'x86_64+fma-centos7-gcc49-opt'}),
      frozenset({'x86_64+avx-slc6-gcc62-opt'}),
      frozenset({'x86_64+avx2-centos7-gcc62-opt'})},
     'x86_64_v3-centos7'),
    ({frozenset({'x86_64+avx2+fma-centos7-gcc49-opt'}),
      frozenset({'x86_64+avx-slc6-gcc62-opt'}),
      frozenset({'x86_64+avx2+fma-centos7-gcc62-opt'})},
     'x86_64_v3-centos7'),
    ({frozenset({'x86_64+avx2+fma-centos7-gcc49-opt'}),
      frozenset({'x86_64+avx-slc6-gcc7-opt'}),
      frozenset({'x86_64+avx2+fma-centos7-gcc62-opt'})},
     'x86_64_v3-centos7'),
    ({frozenset({'x86_64+fma-centos7-gcc7-opt'}),
      frozenset({'x86_64+avx-slc6-gcc62-opt'})},
     'x86_64_v3-centos7'),
    ({frozenset({'x86_64-slc5-gcc49-opt'}),
      frozenset({'x86_64+fma-centos7-gcc7-opt'})},
     'x86_64_v3-any'),
    ({frozenset({'x86_64-slc5-gcc49-opt'}),
      frozenset({'x86_64+fma-centos7-gcc7-opt', 'x86_64+fma-slc6-gcc7-opt'})},
     'x86_64_v3-slc6'),
    ({frozenset({'x86_64-slc5-gcc49-opt', 'x86_64-slc6-gcc49-opt'}),
      frozenset({'x86_64+fma-centos7-gcc7-opt', 'x86_64-slc6-gcc7-opt'})},
     'x86_64_v2-slc6'),
])
def test_getPlatformForJob(binaryTagsSet, expected):
  fbtMock = mock.MagicMock()
  fbtMock.return_value = binaryTagsSet
  moduleTested._findBinaryTags = fbtMock

  res = moduleTested.getPlatformForJob(mock.MagicMock())
  assert res == expected


@pytest.mark.parametrize("platform, expectedRes, expectedValue", [
    ('', True, []),
    ('ANY', True, []),
    (['ANY'], True, []),
    ([None], True, []),
    (['bih', 'boh'], False, []),
    ('x86_64-slc6', True, ['cannonlake-centos7', 'skylake_avx512-centos7', 'skylake-centos7', 'broadwell-centos7',
                           'haswell-centos7', 'ivybridge-centos7', 'sandybridge-centos7', 'westmere-centos7',
                           'nehalem-centos7', 'core2-centos7', 'x86_64-centos7', 'cannonlake-slc6',
                           'skylake_avx512-slc6', 'skylake-slc6', 'broadwell-slc6', 'haswell-slc6', 'ivybridge-slc6',
                           'sandybridge-slc6', 'westmere-slc6', 'nehalem-slc6', 'core2-slc6', 'x86_64-slc6']),
    ('x86_64-centos7', True, ['cannonlake-centos7', 'skylake_avx512-centos7', 'skylake-centos7', 'broadwell-centos7',
                              'haswell-centos7', 'ivybridge-centos7', 'sandybridge-centos7', 'westmere-centos7',
                              'nehalem-centos7', 'core2-centos7', 'x86_64-centos7']),
    ('x86_64-slc5', True, ['cannonlake-slc6', 'skylake_avx512-slc6', 'skylake-slc6', 'broadwell-slc6', 'haswell-slc6',
                           'ivybridge-slc6', 'sandybridge-slc6', 'westmere-slc6', 'nehalem-slc6', 'core2-slc6',
                           'x86_64-slc6', 'cannonlake-slc5', 'skylake_avx512-slc5', 'skylake-slc5', 'broadwell-slc5',
                           'haswell-slc5', 'ivybridge-slc5', 'sandybridge-slc5', 'westmere-slc5', 'nehalem-slc5',
                           'core2-slc5', 'x86_64-slc5']),
    ('x86_64-slc5.avx2', True, ['cannonlake-slc6', 'skylake_avx512-slc6', 'skylake-slc6', 'broadwell-slc6',
                                'haswell-slc6', 'cannonlake-slc5', 'skylake_avx512-slc5', 'skylake-slc5',
                                'broadwell-slc5', 'haswell-slc5']),
    ('x86_64-slc6.avx2+fma', True, ['cannonlake-centos7', 'skylake_avx512-centos7', 'skylake-centos7',
                                    'broadwell-centos7', 'haswell-centos7', 'cannonlake-slc6', 'skylake_avx512-slc6',
                                    'skylake-slc6', 'broadwell-slc6', 'haswell-slc6']),
])
def test_getDIRACPlatform(platform, expectedRes, expectedValue):
  res = moduleTested.getDIRACPlatform(platform)
  assert res['OK'] is expectedRes
  if res['OK']:
    assert set(expectedValue).issubset(set(res['Value'])) is True


@pytest.mark.parametrize("applicationName, applicationVersion, expected", [
    ('DaVinci', 'v44r4', ['x86_64+avx2+fma-centos7-gcc62-opt', 'x86_64-centos7-gcc62-dbg', 'x86_64-centos7-gcc62-do0',
                          'x86_64-centos7-gcc62-opt', 'x86_64-centos7-gcc7-dbg', 'x86_64-centos7-gcc7-opt',
                          'x86_64-slc6-gcc62-dbg', 'x86_64-slc6-gcc62-opt']),
    ('DaVinci', 'v0r0', None),
    ('InvalidProject', 'v1r0', None),
    ('CASTELAO', 'v1r5p1', ['x86_64-centos7-gcc62-do0', 'x86_64-centos7-gcc62-opt', 'x86_64-centos7-gcc62-dbg',
                            'x86_64-slc6-gcc62-opt', 'x86_64-centos7-gcc7-dbg', 'x86_64+avx2+fma-centos7-gcc62-opt',
                            'x86_64-slc6-gcc62-dbg', 'x86_64-centos7-gcc7-opt']),
])
def test_listPlatforms(applicationName, applicationVersion, expected):
  if not os.path.isdir('/cvmfs/lhcb.cern.ch'):
    pytest.skip('CVMFS is required')

  # Good cache path
  result = moduleTested._listPlatforms(
      applicationName, applicationVersion,
      moduleTested.DEFAULT_CACHEPATH)
  assert result is None is expected or set(result) == set(expected)

  # Invalid cache path
  result = moduleTested._listPlatforms(
      applicationName, applicationVersion,
      '/cvmfs/lhcb.cern.invalid/lib/var/lib/softmetadata/project-platforms.json')
  assert result is None



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Agent.NagiosTopologyAgent.

NagiosTopologyAgent.__bases__: DIRAC.Core.Base.AgentModule.AgentModule

xml_append
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time
import xml.dom.minidom
import json
import socket

from six.moves.urllib.request import urlopen
from DIRAC import S_OK, rootPath, gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Storage.StorageElement import StorageElement

__RCSID__ = "$Id$"
AGENT_NAME = 'ResourceStatus/NagiosTopologyAgent'


MAPPING_CE_TYPE = {'lcg': 'CE',
                   'cream': 'CREAM-CE',
                   'arc': 'ARC-CE',
                   'htcondorce': 'HTCONDOR-CE',
                   'vac': 'VAC',
                   'cloud': 'CLOUD',
                   'boinc': 'BOINC',
                   'vcycle': 'VCYCLE',
                   'dirac': 'DIRAC'}


class NagiosTopologyAgent(AgentModule):
  """This agent loops over the Dirac CS and extracts the necessary information
  to create a "topology map" which is used by the IT provided Nagios system to
  test Grid sites. The topology information defines the services to be tested.

  NagiosTopologyAgent, writes the xml topology consumed by Nagios to run
  the tests.
  """

  def __init__(self, *args, **kwargs):

    AgentModule.__init__(self, *args, **kwargs)

    self.xmlPath = 'webRoot/www/topology/'
    self.urljson = 'http://wlcg-cric.cern.ch/api/core/rcsite/query/?json'

    self.dryRun = False

  def initialize(self):
    """Initialize the agent."""

    self.xmlPath = rootPath + '/' + self.am_getOption('webRoot', self.xmlPath)

    try:
      os.makedirs(self.xmlPath)
    except OSError:
      pass  # The dir exists already, or cannot be created: do nothing

    return S_OK()

  def execute(self):
    """ Generates xml and json topology files
    """

    res = getSites()
    if not res['OK']:
      return res
    sites = res['Value']

    # First, create the JSON topology, which contains also what the XML topology needs
    res = self.createJSONTopology(sites)
    if not res['OK']:
      self.log.error("Failed to create JSON topology", res['Message'])
      return res
    self.log.info("Created JSON topology")

    res = self.createXMLTopology(sites)
    if not res['OK']:
      self.log.error("Failed to create XML topology", res['Message'])
      return res
    self.log.info("Created XML topology")

    return S_OK()

  def createJSONTopology(self, sites):
    """ Function that creates a topology.json file, based on specification from WLCG group
    """

    fullSitesDict = dict()
    for site in sites:
      grid = site.split(".")[0]
      res = gConfig.getOptionsDict('Resources/Sites/%s/%s' % (grid, site))
      if not res['OK']:
        self.log.error("Failure getting options dict for site", "%s: %s" % (site, res['Message']))
        continue
      siteInfoCS = res['Value']
      wlcgName = siteInfoCS.get('Name')

      siteInfo = dict()
      siteInfo['WLCG_site'] = wlcgName
      # Tier level 4 is what they want for opportunistic resource
      siteInfo['Tier'] = siteInfoCS.get('MoUTierLevel', '4')
      siteInfo['Type'] = 'GRID' if grid == 'LCG' else site.split('.')[0]

      # Services is for CEs and SEs
      siteInfoServices = dict()

      # CEs
      res = gConfig.getSections(cfgPath('Resources', 'Sites', grid, site, 'CEs'), [])
      if not res['OK']:
        self.log.error("Failure getting CEs section for site", "%s: %s" % (site, res['Message']))
        continue
      if not res['Value']:
        self.log.warn("Site without CEs", "(%s)" % site)
      else:
        cesList = res['Value']
        for ce in cesList:
          res = gConfig.getOptionsDict(cfgPath('Resources', 'Sites', grid, site, 'CEs', ce))
          if not res['OK']:
            self.log.error("Failure getting info on ce", "%s: %s" % (ce, res['Message']))
            continue

          ceDetailsInCS = res['Value']
          ceDetails = dict()
          ceDetails['type'] = 'compute'
          ceDetails['access_points'] = dict()
          ceDetails['access_points'][ce] = dict()
          ceDetails['access_points'][ce]['endpoint_url'] = ce
          ceDetails['access_points'][ce]['type'] = MAPPING_CE_TYPE.get(ceDetailsInCS['CEType'].lower(), 'UNDEFINED')
          ceDetails['access_points'][ce]['monitored'] = 'yes' if wlcgName else 'no'
          ceDetails['access_points'][ce]['quality_level'] = 'production'
          # queues
          res = gConfig.getSections(cfgPath('Resources', 'Sites', grid, site, 'CEs', ce, 'Queues'), [])
          if not res['OK']:
            self.log.error("Failure getting info on ce/queues", "%s: %s" % (ce, res['Message']))
            continue
          if not res['Value']:
            self.log.warn("CE without queues", "(%s)" % ce)
          else:
            for queue in res['Value']:
              ceDetails['compute_shares'] = dict()
              ceDetails['compute_shares'][queue] = dict()
              ceDetails['compute_shares'][queue]['type'] = 'queue'
              ceDetails['compute_shares'][queue]['name'] = queue

          siteInfoServices[ce] = ceDetails

      # SEs
      ses = gConfig.getValue(cfgPath('Resources', 'Sites', grid, site, 'SE'), [])
      if not ses:
        pass
      for se in ses:
        diracSE = StorageElement(se)  # this 'se' is .e.g. 'CERN-DST-EOS'
        seDetailsForDIRACSE = dict()
        seStorageSharesForDIRACSE = dict()
        for diracSEoption in diracSE.protocolOptions:

          # Building the storage_endpoints section
          seDetails = dict()
          ep = os.path.join(
              diracSEoption['Protocol'] + '://' + diracSEoption['Host'].strip('/') + ':' + diracSEoption['Port'],
              diracSEoption['Path'].strip('/'),
          )
          seDetails['endpoint_url'] = ep
          seDetails['interface_type'] = diracSEoption['Protocol']
          seDetails['monitored'] = 'yes' if wlcgName else 'no'
          seDetails['quality_level'] = 'production'

          dseep = '_'.join([se, diracSEoption['Protocol']])  # just a unique name e.g. 'CERN-DST-EOS_root'
          seDetailsForDIRACSE[dseep] = seDetails

          if diracSEoption.get('SpaceToken'):
            # Building the storage_shares section
            seDetails = dict()
            seDetails['Name'] = diracSEoption['SpaceToken']
            seDetails['path'] = diracSEoption['WSUrl']
            if diracSEoption['Protocol'].lower() == 'srm':
              seDetails['assigned_endpoints'] = ep

            dsess = '_'.join([se, diracSEoption['SpaceToken']])  # just a unique name e.g. 'CNAF-DST_LHCb-Disk'
            seStorageSharesForDIRACSE[dsess] = seDetails

        siteInfoServices[se] = dict()
        siteInfoServices[se]['storage_endpoints'] = seDetailsForDIRACSE
        if seStorageSharesForDIRACSE:
          siteInfoServices[se]['storage_shares'] = seStorageSharesForDIRACSE

        siteInfoServices[se]['type'] = 'storage'
        siteInfoServices[se]['quality_level'] = 'production'
        siteInfoServices[se]['monitored'] = 'yes' if wlcgName else 'no'

      siteInfo['Services'] = siteInfoServices
      fullSitesDict[site] = siteInfo

    with open(self.xmlPath + 'topology.json', 'w') as tj:
      json.dump(fullSitesDict, tj)
      self.log.verbose("JSON topology saved", self.xmlPath + 'topology.json')

    return S_OK()

  def createXMLTopology(self, sites):
    """ Creates a lhcb_topology_Generated.xml file to be used for Nagios tests

        :params list sites: list of sites (from DIRAC CS)
    """

    # instantiate xml doc
    xml_impl = xml.dom.minidom.getDOMImplementation()
    xml_doc = xml_impl.createDocument(None, 'root', None)
    xml_root = xml_doc.documentElement

    # xml header info
    writeHeaderInfo(xml_doc, xml_root)

    # loop over sites
    response = urlopen(self.urljson)
    wlcg = json.loads(response.read())

    organized_list_of_sites = []
    for site in sites:
      grid, real_site_name, country = site.split(".")
      same_site = [s for s in sites if ("." + real_site_name + "." + country) in s]
      organized_list_of_sites = organized_list_of_sites + [same_site]

    for site in organized_list_of_sites:
      site_parameters = getSiteParameters(site, wlcg)

      if site_parameters:

        xml_site = xml_append(xml_doc, xml_root, 'atp_site',
                              infrast=site_parameters['Infrastructure'],
                              longitude=site_parameters['Coordinates'].split(":")[0],
                              latitude=site_parameters['Coordinates'].split(":")[1],
                              name=site_parameters['WlcgName'])
        has_grid_elem = False

        dirac_name = site_parameters['DiracName']
        site_tier = site_parameters['Tier']
        site_subtier = site_parameters['Sub-Tier']

        for grid in site_parameters['Grid']:
          site = site_parameters['Grid'][grid]['SiteName']
          ces = site_parameters['Grid'][grid]['CEs']
          # CE info
          if ces:
            res = writeCEInfo(xml_doc, grid, xml_site, site, ces)
            # Update has_grid_elem
            has_grid_elem = res or has_grid_elem

        # SE info
        if site_parameters['SE'] and (site_tier in ['0', '1', '2'] or site_subtier in ['T2-D']):
          # res = self.writeSEInfo( xml_doc, xml_site, dirac_name )
          res = writeSEInfo(xml_doc, xml_site, dirac_name, site_tier, site_subtier)
          # Update has_grid_elem
          has_grid_elem = res or has_grid_elem

        if has_grid_elem:
          xml_append(xml_doc, xml_site, 'group',
                     name='Tier ' + site_tier, type='LHCb_Tier')
          xml_append(xml_doc, xml_site, 'group',
                     name=dirac_name, type='LHCb_Site')
          xml_append(xml_doc, xml_site, 'group',
                     name=dirac_name, type='All Sites')
          xml_append(xml_doc, xml_site, 'group',
                     name=dirac_name, type='WLCG_%s' % (site_parameters['FederationAccountingName']))
          xml_append(xml_doc, xml_site, 'group',
                     name=dirac_name, type='WLCG_%s' % (site_parameters['Country']))
          try:
            if site_subtier == 'T2-D':
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 0/1/2D')
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 2D')
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 2')    # Mod for T2-D also have T2 tag

            elif int(site_tier) == 2:
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 2')

            # site_tier can be only 1 or 0, (see site_tier def above to
            # convince yourself.)
            else:
              # If site_type is None, then we go to the exception.
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 0/1/2D')
              xml_append(xml_doc, xml_site, 'group',
                         name=dirac_name, type='Tier 0/1')

          except ValueError:  # Site tier is None, do nothing
            pass

        else:
          _msg = "Site %s, (WLCG Name: %s) has no CE, SE or LFC, thus will not be put into the xml"
          _msg = _msg % (site, site_parameters['WlcgName'])
          self.log.warn(_msg)
          xml_root.removeChild(xml_site)

    self.dryRun = self.am_getOption('DryRun', self.dryRun)
    if self.dryRun:
      self.log.info("Dry Run: XML file will not be created, just printed")
      print(xml_doc.toxml())

    else:
      # produce the xml
      with open(self.xmlPath + "lhcb_topology_Generated.xml", 'w') as xmlf:
        xmlf.write(xml_doc.toxml())

      self.log.info("XML file created Successfully")

    return S_OK()

# Private methods #######################################################


def isHostIPV6(host):
  """Test if the given host is ipv6 capable.

  0:ipv6 capable. 1:ipv4 only. -1:Not a valid host (no DNS record?)
  """
  try:  # First try IPV6
    socket.getaddrinfo(host, None, socket.AF_INET6)
    return 0
  except socket.gaierror:  # No IPv6 address
    try:  # Next try IPv4
      socket.getaddrinfo(host, None, socket.AF_INET)
      return 1
    except socket.gaierror:  # The host does not exist (no IPv6 or IPv4 address)
      return -1


def getSiteParameters(sites, wlcg):
  """Function that returns the sites parameters.

  :param list sites:
    List of sites or single site with same site name (e.g ['LCG.CERN.cern'] or
    [LCG.Manchester.uk, VAC.Manchester.uk])

  :param dict wlcg:
    It's a dictionary with the WLCG parameters from all sites grabbed from
    http://wlcg-cric.cern.ch/api/core/rcsite/query/?json

  Keys:
  'WlcgName', 'Coordinates', 'Description', 'Mail', 'DiracName', 'Tier', 'Sub-Tier',
  'SE', 'Country':, 'Federation', 'FederationAccountingName', 'Infrastructure',
  'Institute Name', 'Grid'

  If the site is not listed in WLCG or have no MoU Tier Level the function will return False
  """

  grid_dict = {}
  grid, real_site_name, country = sites[0].split(".")
  res = gConfig.getOptionsDict('Resources/Sites/%s/%s' % (grid, sites[0]))
  if not res['OK']:
    gLogger.error("Could not get options", "for site %s: %s" % (sites[0], res['Message']))
    return False
  site_opts = res['Value']
  site_name = site_opts.get('Name')  # This is Grid name (the one in GocDB)
  site_tier = site_opts.get('MoUTierLevel')
  if site_tier and site_name:
    wlcg_params = wlcg.get(site_name)
    if not wlcg_params:
      return False
    if len(sites) > 1:
      for i in sites:
        grid = i.split(".")[0]
        res = gConfig.getSections('Resources/Sites/%s/%s/CEs' % (grid, i))
        grid_dict.update({grid: {'SiteName': i, 'CEs': res['Value'] if res['OK'] else None}})
    else:
      res = gConfig.getSections('Resources/Sites/%s/%s/CEs' % (grid, sites[0]))
      grid_dict.update({grid: {'SiteName': sites[0], 'CEs': res['Value'] if res['OK'] else None}})

    site_subtier = site_opts.get('SubTier', None)
    ses = site_opts.get('SE', None)

    site_params = {'WlcgName': site_opts.get('Name'),
                   'Coordinates': site_opts.get('Coordinates'),
                   'Description': site_opts.get('Description'),
                   'Mail': site_opts.get('Mail'),
                   'DiracName': ('LCG.' + real_site_name + "." + country),
                   'Tier': site_tier,
                   'Sub-Tier': site_subtier,
                   'SE': ses,
                   'Country': wlcg_params.get('country'),
                   'Federation': wlcg_params.get('federations')[0],
                   'FederationAccountingName': wlcg_params.get('federation_accounting_name', site_name),
                   'Infrastructure': wlcg_params.get('infrastructure', 'EGI'),
                   'Institute Name': wlcg_params.get('institute'),
                   'Grid': grid_dict}

    return site_params
  else:
    return False


def writeHeaderInfo(xml_doc, xml_root):
  """Writes XML document header."""

  xml_append(xml_doc, xml_root, 'title', 'LHCb Topology Information for ATP')
  xml_append(xml_doc, xml_root, 'description',
             'List of LHCb site names for monitoring and mapping to the SAM/WLCG site names')
  xml_append(xml_doc, xml_root, 'feed_responsible',
             dn='/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=roiser/CN=564059/CN=Stefan Roiser',
             name='Stefan Roiser')
  xml_append(xml_doc, xml_root, 'last_update',
             time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime()))
  xml_append(xml_doc, xml_root, 'vo', 'lhcb')


def writeCEInfo(xml_doc, grid, xml_site, site, ces):
  """Writes CE information in the XML Document."""

  has_grid_elem = 'False'

  for site_ce_name in ces:

    has_grid_elem = True

    # FIXME: use proper helpers
    site_ce_opts = gConfig.getOptionsDict(
        'Resources/Sites/%s/%s/CEs/%s' % (grid, site, site_ce_name))
    if not site_ce_opts['OK']:
      gLogger.error(site_ce_opts['Message'])
      continue
    site_ce_opts = site_ce_opts['Value']

    site_ce_type = site_ce_opts.get('CEType')

    xml_ce = xml_append(xml_doc, xml_site, 'service', hostname=site_ce_name,
                        flavour=MAPPING_CE_TYPE.get(site_ce_type.lower(), 'UNDEFINED'))

    #   I'll leave this code commented in case it needs to be used in the future,
    #   this function consumes a hell lot of time to return a value that is not
    #   mandatory at the momment
    # ce_batch = ldapCEState(site_ce_name, site_ce_opts['VO'])
    # ce_batch = ce_batch['Value'][0]['GlueCEInfoJobManager'] if (ce_batch['OK'] and ce_batch['Value']) else None

    # ipv6 status of the CE
    i6Status = isHostIPV6(site_ce_name)
    i6Comment = ""
    if i6Status == -1:
      i6Comment = "Maybe DIRAC Service, not a valid machine"
    xml_append(xml_doc, xml_ce, 'queues', ipv6_status=str(i6Status), ipv6_comment=i6Comment)

    ce_queues = gConfig.getSections(
        'Resources/Sites/%s/%s/CEs/%s/Queues/' % (grid, site, site_ce_name))
    if not ce_queues['OK']:
      continue

    ce_queues = ce_queues['Value']
    for queue in ce_queues:
      queue_information = gConfig.getOptionsDict(
          'Resources/Sites/%s/%s/CEs/%s/Queues/%s' % (grid, site, site_ce_name, queue))
      if queue_information['OK']:
        queue_information = queue_information['Value']
        # if queue_information.get('maxCPUTime') > max_CPU:
        #   etf_default = 'True'
        #   max_CPU = queue_information.get('maxCPUTime')

      xml_append(xml_doc, xml_ce, 'queues', ce_resource=queue,
                 # - the following items are already predicter for latter changes if necessary
                 # batch_system=ce_batch,
                 # queue=queue_information.get('VO'),
                 # etf_default=etf_default # => this needs to be fixed, when necessary
                 # maxWaitingJobs=queue_information.get('MaxWaitingJobs'),
                 # maxCPUTime=queue_information.get('maxCPUTime')
                 )
  return has_grid_elem


def writeSEInfo(xml_doc, xml_site, site, site_tier, site_subtier):
  """Writes SE information in the XML Document."""
  def write_SE_XML(site_se_opts):
    """Sub-function just to populate the XML with the SE values."""
    site_se_name = site_se_opts.get('Host')
    site_se_flavour = site_se_opts.get('Protocol')
    site_se_path = site_se_opts.get('Path', 'UNDEFINED')
    site_se_endpoint = site_se_opts.get('URLBase')
    mappingSEFlavour = {'srm': 'SRMv2',
                        'root': 'XROOTD', 'http': 'HTTPS'}

    xml_se = xml_append(xml_doc, xml_site, 'service',
                        endpoint=site_se_endpoint,
                        flavour=mappingSEFlavour.get(site_se_flavour, 'UNDEFINED'),
                        hostname=site_se_name,
                        path=site_se_path)

    # ipv6 status of the SE
    i6Status = isHostIPV6(site_se_name)
    i6Comment = ""
    if i6Status == -1:
      i6Comment = "Maybe DIRAC Service, not a valid machine"
    xml_append(xml_doc, xml_se, 'queues', ipv6_status=str(i6Status), ipv6_comment=i6Comment)

  has_grid_elem = True

  real_site_name = site.split(".")[1]
  dmsHelper = DMSHelpers()

  if int(site_tier) in (0, 1):
    dst = dmsHelper.getSEInGroupAtSite('Tier1-DST', real_site_name)
    raw = dmsHelper.getSEInGroupAtSite('Tier1-RAW', real_site_name)
    if not raw['OK']:
      gLogger.error(raw['Message'])
      return False
    raw = raw['Value']
    se_RAW = StorageElement(raw)
    se_plugins_RAW = se_RAW.getPlugins()

  if site_subtier == 'T2-D':
    dst = dmsHelper.getSEInGroupAtSite('Tier2D-DST', real_site_name)

  if not dst['OK']:
    gLogger.error(dst['Message'])
    return False

  dst = dst['Value']
  se_DST = StorageElement(dst)
  se_plugins_DST = se_DST.getPlugins()
  if not se_plugins_DST['OK']:
    gLogger.error(se_plugins_DST['Message'])
    return False

  for protocol in se_plugins_DST['Value']:
    site_se_opts_DST = se_DST.getStorageParameters(protocol)
    if not site_se_opts_DST['OK']:
      gLogger.error(site_se_opts_DST['Message'])
      return False
    site_se_opts_DST = site_se_opts_DST['Value']
    write_SE_XML(site_se_opts_DST)

    if int(site_tier) in (0, 1):
      if protocol in se_plugins_RAW['Value']:
        site_se_opts_RAW = se_RAW.getStorageParameters(protocol)
        if not site_se_opts_RAW['OK']:
          gLogger.error(site_se_opts_RAW['Message'])
          return has_grid_elem
        site_se_opts_RAW = site_se_opts_RAW['Value']
        # This tests if the DST and RAW StorageElements have the same endpoint.
        # If so it only uses the one already added.
        if site_se_opts_RAW['Host'] != site_se_opts_DST['Host']:
          write_SE_XML(site_se_opts_RAW)

  return has_grid_elem


def xml_append(doc, base, elem, cdata=None, **attrs):
  """Given a Document, we append to it an element."""

  new_elem = doc.createElement(elem)
  for attr in attrs:
    new_elem.setAttribute(attr, attrs[attr])
  if cdata:
    new_elem.appendChild(doc.createTextNode(cdata))

  return base.appendChild(new_elem)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test Interfaces API DiracProduction."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

import LHCbDIRAC.Interfaces.API.DiracProduction as moduleTested
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob

lj = LHCbJob()


def test_LJ_setApplication():
  open('optionsFiles', 'a').close()
  res = lj.setApplication('appName', 'v1r0', 'optionsFiles', systemConfig='x86_64-slc6-gcc-44-opt')
  assert res['OK'] is True
  res = lj.setApplication('appName', 'v1r0', 'optionsFiles', systemConfig='x86_64-slc5-gcc-41-opt')
  assert res['OK'] is True
  res = lj.setApplication('appName', 'v1r0', 'optionsFiles', systemConfig='x86_64-slc5-gcc-43-opt')
  assert res['OK'] is True
  os.remove('optionsFiles')


def test_instantiate():
  """tests that we can instantiate one object of the tested class."""
  testClass = moduleTested.DiracProduction
  prod = testClass(1)
  assert 'DiracProduction' == prod.__class__.__name__



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The ErrorLogging module is used to perform error analysis using AppConfig
utilities. This occurs at the end of each workflow step such that the
step_commons dictionary can be utilized.

Since not all projects are instrumented to work with the AppConfig error
suite any failures will not be propagated to the workflow.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR, gLogger

import LHCbDIRAC.Core.Utilities.LogErr as LogErr
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class ErrorLogging(ModuleBase):

  #############################################################################

  def __init__(self):
    """c'tor."""

    self.log = gLogger.getSubLogger("ErrorLogging")
    super(ErrorLogging, self).__init__(self.log)

    self.version = __RCSID__
    # Internal parameters
    self.errorLogNameHTML = ''
    self.errorLogNamejson = ''

  #############################################################################

  def _resolveInputVariables(self):
    """By convention the module input parameters are resolved here."""
    super(ErrorLogging, self)._resolveInputVariables()
    super(ErrorLogging, self)._resolveInputStep()

    self.errorLogNamejson = '%d_Errors_%s.json' % (self.jobID, self.applicationName)

  #############################################################################

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None):
    """Main execution function. Always return S_OK() because we don't want the
    job execution result to depend on retrieving errors from logs.

    This module will run regardless of the workflow status.
    """

    try:

      super(ErrorLogging, self).execute(self.version, production_id, prod_job_id, wms_job_id,
                                        workflowStatus, stepStatus,
                                        wf_commons, step_commons, step_number, step_id)

      self._resolveInputVariables()

      if self.applicationName.lower() not in ('gauss', 'boole'):
        self.log.info('Not Gauss nor Boole, exiting')
        return S_OK()

      self.log.info('Executing ErrorLogging module for: %s %s %s' % (self.applicationName,
                                                                     self.applicationVersion,
                                                                     self.applicationLog))
      if not os.path.exists(self.applicationLog):
        self.log.info('Application log file from previous module not found locally: %s' % self.applicationLog)
        return S_OK()

      # Now really running
      # self.step_commons['extraPackages'] is something like 'AppConfig.v3r360;TurboStreamProd.v4r2p9;ProdConf'
      appConfigVersion = [x.split('.')[1] for x in self.step_commons['extraPackages'].split(';') if 'AppConfig' in x][0]
      result = LogErr.readLogFile(
          logFile=self.applicationLog,
          project=self.applicationName,
          version=self.applicationVersion,
          appConfigVersion=appConfigVersion,
          jobID=self.prod_job_id,
          prodID=self.production_id,
          wmsID=self.jobID,
          name=self.errorLogNamejson)

      if not result['OK']:
        self.log.error("Error logging for %s %s step %s completed with errors:" % (self.applicationName,
                                                                                   self.applicationVersion,
                                                                                   self.step_number))
        self.log.warn('Exiting without affecting workflow status')
        return S_OK()

      if not os.path.exists(self.errorLogNamejson):
        self.log.error('%s not found locally, exiting without affecting workflow status' % self.errorLogNamejson)
        return S_OK()

      self.log.info("Error logging for %s %s step %s completed successfully:" % (self.applicationName,
                                                                                 self.applicationVersion,
                                                                                 self.step_number))

      return S_OK()

    except Exception as e:
      self.log.exception("Failure in ErrorLogging execute module", lException=e)
      return S_ERROR("Error in ErrorLogging module")

    finally:
      super(ErrorLogging, self).finalize(self.version)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit test of ConsistencyChecks."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=invalid-name,missing-docstring,protected-access

import unittest
from mock import Mock

from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks


class UtilitiesTestCase(unittest.TestCase):

  def setUp(self):

    self.dmMock = Mock()
    self.dmMock.getReplicas.return_value = {'OK': True, 'Value': {'Successful': {'bb.raw': 'metadataPippo'},
                                                                  'Failed': {}}}

    self.cc = ConsistencyChecks(transClient=Mock(), dm=self.dmMock, bkClient=bkc_mock)
    self.fileTypes = [['SEMILEPTONIC.DST', 'LOG', 'RAW'], [
        'SEMILEPTONIC.DST', 'LOG', 'RAW'], ['SEMILEPTONIC.DST'], ['SEMILEPTONIC.DST']]
    self.cc.fileTypesExcluded = ['LOG']
    self.cc.prod = 0
    self.maxDiff = None

  def tearDown(self):
    pass


class ConsistencyChecksSuccess(UtilitiesTestCase):

  def test__selectByFileType(self):

    lfnDicts = [{'aa.raw': {'bb.raw': {'FileType': 'RAW', 'RunNumber': 97019},
                            'bb.log': {'FileType': 'LOG'},
                            '/bb/pippo/aa.dst': {'FileType': 'DST'},
                            '/lhcb/1_2_1.Semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}},
                 'cc.raw': {'dd.raw': {'FileType': 'RAW', 'RunNumber': 97019},
                            'bb.log': {'FileType': 'LOG'},
                            '/bb/pippo/aa.dst': {'FileType': 'LOG'},
                            '/lhcb/1_1.semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}}
                 },
                {'aa.raw': {'/bb/pippo/aa.dst': {'FileType': 'LOG'},
                            'bb.log': {'FileType': 'LOG'}}
                 },
                {'aa.raw': {'bb.raw': {'FileType': 'RAW', 'RunNumber': 97019},
                            'bb.log': {'FileType': 'LOG'},
                            '/bb/pippo/aa.dst': {'FileType': 'DST'},
                            '/lhcb/1_2_1.Semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}},
                 'cc.raw': {'dd.raw': {'FileType': 'RAW', 'RunNumber': 97019},
                            'bb.log': {'FileType': 'LOG'},
                            '/bb/pippo/aa.dst': {'FileType': 'LOG'},
                            '/lhcb/1_1.semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}}
                 },
                {'aa.raw': {'/bb/pippo/aa.dst': {'FileType': 'LOG'},
                            'bb.log': {'FileType': 'LOG'}}
                 },
                ]

    lfnDictsExpected = [{'aa.raw':
                         {'/lhcb/1_2_1.Semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'},
                          'bb.log': {'FileType': 'LOG'},
                          'bb.raw': {'RunNumber': 97019, 'FileType': 'RAW'}},
                         'cc.raw':
                         {'dd.raw': {'RunNumber': 97019, 'FileType': 'RAW'},
                          'bb.log': {'FileType': 'LOG'},
                             '/bb/pippo/aa.dst': {'FileType': 'LOG'},
                          '/lhcb/1_1.semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}}
                         },
                        {'aa.raw':
                         {'/bb/pippo/aa.dst': {'FileType': 'LOG'},
                          'bb.log': {'FileType': 'LOG'}}
                         },
                        {'aa.raw':
                         {'/lhcb/1_2_1.Semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}},
                         'cc.raw':
                         {'/lhcb/1_1.semileptonic.dst': {'FileType': 'SEMILEPTONIC.DST'}}
                         },
                        {}
                        ]

    # testing various cases
    for fileType, lfnDict, lfnDictExpected in zip(self.fileTypes, lfnDicts, lfnDictsExpected):
      self.cc.fileType = fileType

      res = self.cc._selectByFileType(lfnDict)

      self.assertEqual(res, lfnDictExpected)

      res = self.cc._selectByFileType(lfnDict)
      self.assertEqual(res, lfnDictExpected)

  def test__getFileTypesCount(self):
    lfnDict = {'aa.raw': {'bb.log': {'FileType': 'LOG'},
                          '/bb/pippo/aa.dst': {'FileType': 'DST'}}}
    res = self.cc._getFileTypesCount(lfnDict)
    resExpected = {'aa.raw': {'DST': 1, 'LOG': 1}}
    self.assertEqual(res, resExpected)

    lfnDict = {'aa.raw': {'bb.log': {'FileType': 'LOG'},
                          '/bb/pippo/aa.dst': {'FileType': 'DST'},
                          '/bb/pippo/cc.dst': {'FileType': 'DST'}}}
    res = self.cc._getFileTypesCount(lfnDict)
    resExpected = {'aa.raw': {'DST': 2, 'LOG': 1}}
    self.assertEqual(res, resExpected)

  def test_getDescendants(self):
    self.cc.fileType = ['SEMILEPTONIC.DST', 'LOG', 'RAW']
    res = self.cc.getDescendants(['aa.raw'])
    filesWithDescendants, filesWithoutDescendants, filesWitMultipleDescendants, \
        descendants, inFCNotInBK, inBKNotInFC, removedFiles, inFailover = res
    self.assertEqual(
        {k: set(v) for k, v in filesWithDescendants.items()},
        {'aa.raw': {'bb.log', 'bb.raw'}},
    )
    self.assertEqual(filesWithoutDescendants, {})
    self.assertEqual(filesWitMultipleDescendants, {})
    self.assertEqual(descendants, ['bb.raw'])
    self.assertEqual(inFCNotInBK, [])
    self.assertEqual(inBKNotInFC, [])
    self.assertEqual(removedFiles, [])
    self.assertEqual(inFailover, [])


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ConsistencyChecksSuccess))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Production API.

A production is an augmented version of an LHCbJob

Notes:

- Supports all workflows
- create() method that takes a workflow or Production object
  and publishes to the production management system, in addition this
  can automatically construct and publish the BK pass info and transformations
- Uses __getOutputLFNs() function to add production output directory parameter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import shutil
import re
import os

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Workflow.Workflow import Workflow, fromXMLString
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getSiteTier
from DIRAC.Workflow.Utilities.Utils import getStepDefinition

from LHCbDIRAC.Core.Utilities.ProductionData import preSubmissionLFNs
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation


class Production(object):
  """Production uses an LHCbJob object, as well as few clients."""

  #############################################################################

  def __init__(self, script=None):
    """Instantiates the Workflow object and some default parameters."""

    self.LHCbJob = LHCbJob(script)
    self.bkkClient = BookkeepingClient()
    self.transformation = Transformation()
    self.opsHelper = Operations()

    self.bkSteps = {}
    self.prodGroup = ''
    self.plugin = ''
    self.inputFileMask = ''
    self.inputBKSelection = {}
    self.jobFileGroupSize = 0
    self.ancestorProduction = 0
    self.transformationFamily = 0
    self.priority = 1
    self.gaudiSteps = []

    # a dictionary with 'FileType':'SEName'. All the file types should be present. HISTs are present by default
    histogramSE = self.opsHelper.getValue('Productions/HistogramSE', 'CERN-HIST')
    histoTypes = self.opsHelper.getValue('Productions/HistogramTypes', ['HIST', 'BRUNELHIST', 'DAVINCIHIST',
                                                                        'GAUSSHIST'])
    self.outputSEs = dict((ht, histogramSE) for ht in histoTypes)

    if not script:
      self.__setDefaults()

  #############################################################################

  def __setDefaults(self):
    """Sets some default parameters."""

    self.LHCbJob.stepCount = 0
    self.LHCbJob.setOutputSandbox(self.opsHelper.getValue('Productions/inOutputSandbox',
                                                          ['std.out', 'std.err', '*.log']))
    self.setJobParameters({'Type': 'MCSimulation',
                           'Platform': 'ANY',
                           'CPUTime': '1000000',
                           'LogLevel': 'verbose',
                           'JobGroup': '@{PRODUCTION_ID}'})

    self.setFileMask('')

    # General workflow parameters
    self.setParameter('PRODUCTION_ID', 'string', '00012345', 'ProductionID')
    self.setParameter('JOB_ID', 'string', '00006789', 'ProductionJobID')
    self.setParameter('poolXMLCatName', 'string', 'pool_xml_catalog.xml', 'POOLXMLCatalogName')
    self.setParameter('outputMode', 'string', 'Any', 'SEResolutionPolicy')
    self.setParameter('outputDataFileMask', 'string', '', 'outputDataFileMask')

    # BK related parameters
    self.setParameter('configName', 'string', 'MC', 'ConfigName')
    self.setParameter('configVersion', 'string', '2009', 'ConfigVersion')
    self.setParameter('conditions', 'string', '', 'SimOrDataTakingCondsString')

  #############################################################################

  def setJobParameters(self, parametersDict):
    """Set an (LHCb)Job parameter.

    The parametersDict is in the form {'parameterName': 'value'} Each
    parameter calls LHCbJob.setparameterName(value)
    """

    for parameter in parametersDict:
      getattr(self.LHCbJob, 'set' + parameter)(parametersDict[parameter])

  #############################################################################

  def setParameter(self, name, parameterType, parameterValue, description):
    """Set parameters checking in CS in case some defaults need to be
    changed."""
    proposedParam = self.opsHelper.getValue('Productions/%s' % name, '')
    if proposedParam:
      gLogger.debug('Setting %s from CS defaults = %s' % (name, proposedParam))
      self.LHCbJob._addParameter(self.LHCbJob.workflow, name, parameterType, proposedParam,
                                 description)  # pylint: disable=protected-access
    else:
      gLogger.debug('Setting parameter %s = %s' % (name, parameterValue))
      self.LHCbJob._addParameter(self.LHCbJob.workflow, name, parameterType, parameterValue,
                                 description)  # pylint: disable=protected-access

  #############################################################################

  @staticmethod
  def __checkArguments(extraPackages, optionsFile):
    """Checks for typos in the structure of standard arguments to workflows.

    In case of any non-standard settings will raise an exception
    preventing creation of the production. Must be called after setting
    the first event type of the production.
    """
    if not extraPackages:
      extraPackages = []

    if not optionsFile:
      optionsFile = []

    if extraPackages:
      if not re.search(';', extraPackages):
        extraPackages = [extraPackages]
      else:
        extraPackages = extraPackages.split(';')
    if optionsFile:
      if not re.search(';', optionsFile):
        optionsFile = [optionsFile]

    for package in extraPackages:
      gLogger.verbose("Checking extra package: %s" % (package))
      if not re.search('.', package):
        raise TypeError("Must have extra packages in the following format 'Name.Version' not %s" % (package))

    gLogger.verbose("Extra packages and event type options are correctly specified")
    return S_OK()

  #############################################################################

  def addApplicationStep(self, stepDict, inputData=None, modulesList=None, extraModulesList=None):
    """Adds an application step to the workflow.

    stepDict contains everything that is in the step, for this production, e.g.::

      {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 13718, 'ApplicationVersion': 'v28r3p1',
       'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'ExtraOptions': '',
       'ProcessingPass': 'Merging', 'Visible': 'N', 'OptionsFormat': '',
       'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py',
       'DDDB': 'head-20110302', 'CONDDB': 'head-20110407', 'DQTag': '',
       'isMulticore': 'N', 'SystemConfig': '', 'mcTCK': '',
       'fileTypesIn': ['SDST'],
       'visibilityFlag': [{'Visible': 'Y', 'FileType': 'BHADRON.DST'}],
       'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST', 'CHARM.MDST', 'CHARMCOMPLETEEVENT.DST']}

    Note: this step treated here does not necessarily corresponds to a step of the BKK:
    the case where they might be different is the merging case.

    :param dict stepDict: contains everything that is in the step, for this production
    :param str inputData: the input data of the step. Either None, or 'previousStep', or the input to the step
    :param list modulesList: the list of module names (str) this step is made of. If None, a default is taken.

    :returns: the name (str) of the step added
    """

    appName = stepDict['ApplicationName']
    appVersion = stepDict['ApplicationVersion']
    optionsFile = stepDict['OptionFiles']
    stepID = stepDict['StepId']
    stepName = stepDict['StepName']
    stepVisible = stepDict['Visible']
    extraPackages = stepDict['ExtraPackages']
    optionsLine = stepDict['ExtraOptions']
    fileTypesIn = stepDict['fileTypesIn']
    fileTypesOut = stepDict['fileTypesOut']
    stepPass = stepDict['ProcessingPass']
    optionsFormat = stepDict['OptionsFormat']
    dddbOpt = stepDict['DDDB']
    conddbOpt = stepDict['CONDDB']
    dqOpt = stepDict['DQTag']
    multicore = stepDict['isMulticore']
    sysConfig = stepDict['SystemConfig']
    outputVisibility = stepDict['visibilityFlag']
    if sysConfig == 'None' or sysConfig == 'NULL' or not sysConfig or sysConfig is None:
      sysConfig = 'ANY'
    mcTCK = stepDict['mcTCK']

    if extraPackages:
      if isinstance(extraPackages, list):
        extraPackages = ';'.join(extraPackages)
      if not isinstance(extraPackages, str):
        raise TypeError("extraPackages is not a string (nor a list)")
      if ',' in extraPackages:
        extraPackages = ';'.join(extraPackages.split(','))
      if 'ProdConf' not in extraPackages:
        extraPackages = extraPackages + ';ProdConf'
      extraPackages = extraPackages.replace(' ', '')
    else:
      extraPackages = 'ProdConf'

    self.__checkArguments(extraPackages, optionsFile)

    try:
      if not dddbOpt.lower() == 'global':
        gLogger.verbose("Specific DDDBTag setting found for %s step, setting to: %s" % (appName, dddbOpt))
        dddbOpt = dddbOpt.replace(' ', '')
    except AttributeError:
      pass
    try:
      if not conddbOpt.lower() == 'global':
        gLogger.verbose("Specific CondDBTag setting found for %s step, setting to: %s" % (appName, conddbOpt))
        conddbOpt = conddbOpt.replace(' ', '')
    except AttributeError:
      pass
    try:
      if not dqOpt.lower() == 'global':
        gLogger.verbose("Specific DQTag setting found for %s step, setting to: %s" % (appName, dqOpt))
        dqOpt = dqOpt.replace(' ', '')
    except AttributeError:
      pass

    # starting real stuff
    self.LHCbJob.stepCount += 1

    if 'Gaudi_App_Step' not in self.LHCbJob.workflow.step_definitions:

      gLogger.debug("Determining the modules of the steps (modulesList = %s)" % modulesList)
      if modulesList is None:  # we assume it's a standard list of modules for Gaudi steps
        gaudiPath = 'Productions/GaudiStep_Modules'
        modulesList = self.opsHelper.getValue(gaudiPath, ['GaudiApplication', 'AnalyseXMLSummary',
                                                          'ErrorLogging', 'BookkeepingReport', 'StepAccounting'])
      if extraModulesList:
        modulesList += extraModulesList

      # pName, pType, pValue, pDesc
      parametersList = [['inputData', 'string', '', 'StepInputData'],
                        ['inputDataType', 'string', '', 'InputDataType'],
                        ['applicationName', 'string', '', 'ApplicationName'],
                        ['applicationVersion', 'string', '', 'ApplicationVersion'],
                        ['runTimeProjectName', 'string', '', 'runTimeProjectName'],
                        ['runTimeProjectVersion', 'string', '', 'runTimeProjectVersion'],
                        ['applicationType', 'string', '', 'ApplicationType'],
                        ['optionsFile', 'string', '', 'OptionsFile'],
                        ['extraOptionsLine', 'string', '', 'extraOptionsLines'],
                        ['listoutput', 'list', [], 'StepOutputList'],
                        ['extraPackages', 'string', '', 'ExtraPackages'],
                        ['BKStepID', 'string', '', 'BKKStepID'],
                        ['StepProcPass', 'string', '', 'StepProcessingPass'],
                        ['optionsFormat', 'string', '', 'ProdConf configuration'],
                        ['CondDBTag', 'string', '', 'ConditionDatabaseTag'],
                        ['DDDBTag', 'string', '', 'DetDescTag'],
                        ['DQTag', 'string', '', 'DataQualityTag'],
                        ['multiCore', 'string', '', 'MultiCore Flag'],
                        ['SystemConfig', 'string', '', 'system config'],
                        ['mcTCK', 'string', '', 'TCK to be simulated']]

      gaudiStepDef = getStepDefinition('Gaudi_App_Step', modulesNameList=modulesList,
                                       importLine='LHCbDIRAC.Workflow.Modules',
                                       parametersList=parametersList)
      self.LHCbJob.workflow.addStep(gaudiStepDef)

    # create the step instance add it to the wf, and return it
    name = '%s_%s' % (appName, self.LHCbJob.stepCount)
    gaudiStepInstance = self.LHCbJob.workflow.createStepInstance('Gaudi_App_Step', name)

    valuesToSet = [['applicationName', appName],
                   ['applicationVersion', appVersion],
                   ['optionsFile', optionsFile],
                   ['extraOptionsLine', optionsLine],
                   ['extraPackages', extraPackages],
                   ['BKStepID', str(stepID)],
                   ['StepProcPass', stepPass],
                   ['optionsFormat', optionsFormat],
                   ['CondDBTag', conddbOpt],
                   ['DDDBTag', dddbOpt],
                   ['DQTag', dqOpt],
                   ['multiCore', multicore],
                   ['SystemConfig', sysConfig],
                   ['mcTCK', mcTCK]]

    if fileTypesIn:
      valuesToSet.append(['inputDataType', ';'.join(ftIn.upper() for ftIn in fileTypesIn)])

    if inputData is None:
      gLogger.verbose('%s step has no data requirement or is linked to the overall input data' % appName)
      gaudiStepInstance.setLink('inputData', 'self', 'InputData')
    elif inputData == 'previousStep':
      gLogger.verbose('Taking input data as output from previous Gaudi step')
      valuesToSet.append(['inputData', inputData])
    else:
      gLogger.verbose('Assume input data requirement should be added to job')
      self.LHCbJob.setInputData(inputData)
      gaudiStepInstance.setLink('inputData', 'self', 'InputData')

    for pName, value in valuesToSet:
      if value:
        gaudiStepInstance.setValue(pName, value)

    outputFilesList = self._constructOutputFilesList(fileTypesOut)
    gaudiStepInstance.setValue('listoutput', (outputFilesList))

    # to construct the BK processing pass structure, starts from step '0'
    stepIDInternal = 'Step%s' % (self.LHCbJob.stepCount - 1)
    bkOptionsFile = optionsFile

    stepBKInfo = {'ApplicationName': appName,
                  'ApplicationVersion': appVersion,
                  'OptionFiles': bkOptionsFile,
                  'DDDb': dddbOpt,
                  'CondDb': conddbOpt,
                  'DQTag': dqOpt,
                  'ExtraPackages': extraPackages,
                  'BKStepID': stepID,
                  'StepName': stepName,
                  'StepVisible': stepVisible,
                  'OutputFileTypes': outputVisibility}

    self.bkSteps[stepIDInternal] = stepBKInfo
    self.__addBKPassStep()

    return name

  #############################################################################

  def _constructOutputFilesList(self, filesTypesList):
    """Build list of dictionary of output file types, including HIST case.

    :param list filesTypesList: a list of file types (str)
    :returns: a list with a dictionary of file types, lowered
    """

    outputList = []

    for fileType in filesTypesList:
      fileDict = {}
      fileDict['outputDataType'] = fileType.lower()
      outputList.append(fileDict)

    return outputList

  #############################################################################

  def __addBKPassStep(self):
    """Internal method to add BKK parameters."""
    bkPass = 'BKProcessingPass'
    description = 'BKProcessingPassInfo'
    self.LHCbJob._addParameter(self.LHCbJob.workflow, bkPass, 'dict', self.bkSteps,
                               description)  # pylint: disable=protected-access

  #############################################################################

  def addFinalizationStep(self, modulesList=None):
    """Add the finalization step to the workflow (some defaults are inserted)

    :param list modulesList: the list of modules names this step is made of. If None, a default is taken.
    :returns: None
    """
    if modulesList is None:
      modulesList = ['UploadOutputData', 'UploadLogFile', 'UploadMC', 'FailoverRequest']

    if 'Job_Finalization' not in self.LHCbJob.workflow.step_definitions:

      jobFinalizationStepDef = getStepDefinition('Job_Finalization',
                                                 importLine='LHCbDIRAC.Workflow.Modules',
                                                 modulesNameList=modulesList)
      self.LHCbJob.workflow.addStep(jobFinalizationStepDef)

    # create the step instance add it to the wf
    self.LHCbJob.workflow.createStepInstance('Job_Finalization', 'finalization')

  #############################################################################

  def _lastParameters(self):
    """Add the last parameters before creating the xml file containing the
    workflow."""

    self.LHCbJob._addParameter(self.LHCbJob.workflow, 'gaudiSteps', 'list', self.gaudiSteps,
                               'list of Gaudi Steps')  # pylint: disable=protected-access
    self.LHCbJob._addParameter(self.LHCbJob.workflow, 'outputSEs', 'dict', self.outputSEs,
                               'dictionary of output SEs')  # pylint: disable=protected-access

  def __createWorkflow(self, name=''):
    """Create XML of the workflow."""
    self._lastParameters()

    if not name:
      name = self.LHCbJob.workflow.getName()
    if not re.search('xml$', name):
      name = '%s.xml' % name
    if os.path.exists(name):
      shutil.move(name, '%s.backup' % name)
    name = name.replace('/', '').replace('\\', '')
    self.LHCbJob.workflow.toXMLFile(name)

    return name

  #############################################################################

  def runLocal(self):
    """Create XML workflow for local testing then reformulate as a job and run
    locally."""

    xmlFileName = self.__createWorkflow()
    # it makes a job (a Worklow, with Parameters), out of the xml file
    return LHCbJob(xmlFileName).runLocal()

  #############################################################################

  def __getProductionParameters(self, prodXMLFile, prodID, groupDescription='',
                                bkPassInfo={}, derivedProd=0, reqID=0):
    """This method will publish production parameters."""

    prodWorkflow = Workflow(prodXMLFile)

    parameters = {}
    info = []

    for parameterName in ('Priority', 'CondDBTag', 'DDDBTag', 'DQTag', 'eventType',
                          'processingPass', 'FractionToProcess',
                          'MinFilesToProcess', 'configName', 'configVersion',
                          'outputDataFileMask', 'JobType', 'MaxNumberOfTasks'):
      try:
        parameters[parameterName] = prodWorkflow.findParameter(parameterName).getValue()
        info.append("%s: %s" % (parameterName, prodWorkflow.findParameter(parameterName).getValue()))
      except AttributeError:
        continue

    parameters['SizeGroup'] = self.jobFileGroupSize

    if prodWorkflow.findParameter('InputData'):  # now only comes from BK query
      prodWorkflow.findParameter('InputData').setValue('')
      gLogger.verbose('Resetting input data for production to null, this comes from a BK query...')
      prodXMLFile = self.__createWorkflow(prodXMLFile)
      # prodWorkflow.toXMLFile(prodXMLFile)

    if self.transformationFamily:
      parameters['TransformationFamily'] = self.transformationFamily

    if not bkPassInfo:
      bkPassInfo = prodWorkflow.findParameter('BKProcessingPass').getValue()
    if not groupDescription:
      groupDescription = prodWorkflow.findParameter('groupDescription').getValue()

    parameters['BKCondition'] = prodWorkflow.findParameter('conditions').getValue()
    parameters['BKProcessingPass'] = bkPassInfo
    parameters['groupDescription'] = groupDescription
    parameters['RequestID'] = reqID
    parameters['DerivedProduction'] = derivedProd

    result = self.__getOutputLFNs(prodID, '99999999', prodXMLFile)
    if not result['OK']:
      gLogger.error('Could not create production LFNs', result)

    outputLFNs = result['Value']
    parameters['OutputLFNs'] = outputLFNs

    # the list of output directories is later used for consistency check and for removing output data
    outputDirectories = []
    del outputLFNs['BookkeepingLFNs']  # since ProductionOutputData uses the file mask
    for i in outputLFNs.values():
      for j in i:
        outputDir = '%s%s' % (j.split(str(prodID))[0], prodID)
        if outputDir not in outputDirectories:
          outputDirectories.append(outputDir)

    parameters['OutputDirectories'] = outputDirectories

    # Now for the steps of the workflow
    stepKeys = sorted(bkPassInfo)
    for step in stepKeys:
      info.append('====> %s %s %s' % (bkPassInfo[step]['ApplicationName'],
                                      bkPassInfo[step]['ApplicationVersion'],
                                      step))
      info.append('%s Option Files:' % (bkPassInfo[step]['ApplicationName']))
      if bkPassInfo[step]['OptionFiles']:
        for opts in bkPassInfo[step]['OptionFiles'].split(';'):
          info.append('%s' % opts)
      if bkPassInfo[step]['ExtraPackages']:
        info.append('ExtraPackages: %s' % (bkPassInfo[step]['ExtraPackages']))

    try:
      if parameters['BkQuery']:
        info.append('\nBK Input Data Query:')
        for bkn, bkv in parameters['BkQuery'].items():
          info.append('%s= %s' % (bkn, bkv))
    except KeyError:
      pass

    # BK output directories (very useful)
    bkPaths = []
    bkOutputPath = '%s/%s/%s/%s/%s/%s' % (parameters['configName'],
                                          parameters['configVersion'],
                                          parameters['BKCondition'],
                                          parameters.get('processingPass', ''),
                                          parameters['groupDescription'],
                                          parameters['eventType'])
    fileTypes = parameters['outputDataFileMask']
    fileTypes = [a.upper() for a in fileTypes.split(';')]

    # Annoying that histograms are extension root
    if 'ROOT' in fileTypes:
      fileTypes.remove('ROOT')
      fileTypes.append('HIST')

    for ft in fileTypes:
      bkPaths.append('%s/%s' % (bkOutputPath, ft))
    parameters['BKPaths'] = bkPaths
    info.append('\nBK Browsing Paths:\n%s' % ('\n'.join(bkPaths)))
    infoString = '\n'.join(info)
    parameters['DetailedInfo'] = infoString

    gLogger.verbose('Parameters that will be added: %s' % parameters)

    return parameters

  #############################################################################

  def create(self, publish=True,
             wfString='', requestID=0, reqUsed=0):
    """Will create the production and subsequently publish to the BK.
    Production parameters are also added at this point.

    publish = True - will add production to the production management system
              False - does not publish the production

    The workflow XML is created regardless of the flags.
    """

    if wfString:
      self.LHCbJob.workflow = fromXMLString(wfString)
#      self.name = self.LHCbJob.workflow.getName()

    bkConditions = self.LHCbJob.workflow.findParameter('conditions').getValue()

    bkSteps = self.LHCbJob.workflow.findParameter('BKProcessingPass').getValue()

    bkDictStep = {}

    # Add the BK conditions metadata / name
    simConds = self.bkkClient.getSimConditions()
    if not simConds['OK']:
      gLogger.error('Could not retrieve conditions data from BK:\n%s' % simConds)
      return simConds
    simulationDescriptions = []
    for record in simConds['Value']:
      simulationDescriptions.append(str(record[1]))

    if bkConditions not in simulationDescriptions:
      gLogger.verbose('Assuming BK conditions %s are DataTakingConditions' % bkConditions)
      bkDictStep['DataTakingConditions'] = bkConditions
    else:
      gLogger.verbose('Found simulation conditions for %s' % bkConditions)
      bkDictStep['SimulationConditions'] = bkConditions

    descShort = self.LHCbJob.workflow.getDescrShort()
    descLong = self.LHCbJob.workflow.getDescription()

    prodID = 0
    if publish:

      self.setParameter('ProcessingType', 'JDL', str(self.prodGroup), 'ProductionGroupOrType')
      self.setParameter('Priority', 'JDL', str(self.priority), 'UserPriority')

      try:
        fileName = self.__createWorkflow()
      except Exception as x:  # pylint: disable=broad-except
        gLogger.error(x)
        return S_ERROR('Could not create workflow')

      gLogger.verbose('Workflow XML file name is: %s' % fileName)

      workflowBody = ''
      if os.path.exists(fileName):
        with open(fileName, 'r') as fopen:
          workflowBody = fopen.read()
      else:
        return S_ERROR('Could not get workflow body')

      # Standard parameters
      self.transformation.setTransformationName(fileName)
      self.transformation.setTransformationGroup(self.prodGroup)
      self.transformation.setDescription(descShort)
      self.transformation.setLongDescription(descLong)
      self.transformation.setType(self.LHCbJob.type)
      self.transformation.setBody(workflowBody)
      self.transformation.setPlugin(self.plugin)
      self.transformation.setBkQuery(self.inputBKSelection)
      self.transformation.setTransformationFamily(self.transformationFamily)
      self.transformation.setFileMask(self.inputFileMask)
      self.transformation.setGroupSize(int(self.jobFileGroupSize))
      self.transformation.setInheritedFrom(int(self.ancestorProduction))

      result = self.transformation.addTransformation()

      if not result['OK']:
        gLogger.error('Problem creating production:\n%s' % result)
        return result
      prodID = result['Value']
      gLogger.info('Production %s successfully created' % prodID)

      # All other parameters
      groupDesc = self.LHCbJob.workflow.findParameter('groupDescription').getValue()
      paramsDict = self.__getProductionParameters(prodID=prodID,
                                                  prodXMLFile=fileName,
                                                  groupDescription=groupDesc,
                                                  bkPassInfo=bkSteps,
                                                  reqID=requestID,
                                                  derivedProd=self.ancestorProduction)
      for parName, parValue in paramsDict.items():
        result = getattr(self.transformation, 'set' + parName)(parValue)

    else:
      gLogger.verbose('Publish flag is disabled, using default production ID')

    bkDictStep['Production'] = int(prodID)

    if self.inputBKSelection:
      queryProdID = int(self.inputBKSelection.get('ProductionID', 0))
      queryProcPass = self.inputBKSelection.get(
          'ProcessingPass', '') if self.inputBKSelection.get(
          'ProcessingPass', '') != 'All' else ''

      if queryProdID:
        inputPass = self.bkkClient.getProductionProcessingPass(queryProdID)
        if not inputPass['OK']:
          gLogger.error(inputPass)
          gLogger.error('Production %s was created but BK processing pass for %d was not found' % (prodID,
                                                                                                   queryProdID))
          return inputPass
        inputPass = inputPass['Value']
        gLogger.info('Setting %d as BK input production for %s with processing pass %s' % (queryProdID,
                                                                                           prodID,
                                                                                           inputPass))
        bkDictStep['InputProductionTotalProcessingPass'] = inputPass
      elif queryProcPass:
        gLogger.info('Adding input BK processing pass for production %s from input data query: %s' % (prodID,
                                                                                                      queryProcPass))
        bkDictStep['InputProductionTotalProcessingPass'] = queryProcPass
    else:
      # has to account for an input processing pass anyway, if there is one,
      # or output files may not have the output processing pass correctly computed
      try:
        bkDictStep['InputProductionTotalProcessingPass'] = self.LHCbJob.workflow.findParameter(
            'processingPass').getValue()
      except AttributeError:
        pass

    stepList = []
    stepKeys = sorted(bkSteps)
    # The BK needs an ordered list of steps
    for step in stepKeys:
      stepID = bkSteps[step]['BKStepID']
      if stepID:
        stepName = bkSteps[step]['StepName']
        stepVisible = bkSteps[step]['StepVisible']
        outputFileTypes = bkSteps[step]['OutputFileTypes']
        stepList.append({'StepId': int(stepID),
                         'OutputFileTypes': outputFileTypes,
                         'StepName': stepName,
                         'Visible': stepVisible})

    # This is the last component necessary for the BK publishing (post reorganisation)
    bkDictStep['Steps'] = stepList
    bkDictStep['EventType'] = paramsDict['eventType']

    bkDictStep['ConfigName'] = self.LHCbJob.workflow.findParameter('configName').getValue()
    bkDictStep['ConfigVersion'] = self.LHCbJob.workflow.findParameter('configVersion').getValue()

    if publish:
      gLogger.verbose('Attempting to publish production %s to the BK' % (prodID))
      result = self.bkkClient.addProduction(bkDictStep)
      if not result['OK']:
        gLogger.error(result)
        return result

    if requestID and publish:
      reqDict = {'ProductionID': int(prodID), 'RequestID': requestID, 'Used': reqUsed, 'BkEvents': 0}
      result = ProductionRequestClient(timeout=120).addProductionToRequest(reqDict)
      if not result['OK']:
        gLogger.error('Attempt to add production %s to request %s failed: %s ' % (prodID, requestID,
                                                                                  result['Message']))
        gLogger.error('Dictionary below:\n%s' % reqDict)
      else:
        gLogger.info('Successfully added production %s to request %s with flag set to %s' % (prodID,
                                                                                             requestID,
                                                                                             reqUsed))

    return S_OK(prodID)

  #############################################################################

  def __getOutputLFNs(self, prodID='12345', prodJobID='6789', prodXMLFile=''):
    """Will construct the output LFNs for the production for visual
    inspection."""
    if not prodXMLFile:
      gLogger.verbose('Using workflow object to generate XML file')
      prodXMLFile = self.__createWorkflow()

    job = LHCbJob(prodXMLFile)
    result = preSubmissionLFNs(job._getParameters(), job.workflow.createCode(),  # pylint: disable=protected-access
                               productionID=prodID, jobID=prodJobID)
    if not result['OK']:
      return result
    lfns = result['Value']
    gLogger.verbose(lfns)
    return result

  #############################################################################

  def setFileMask(self, fileMask='', stepMask=''):
    """Output data related parameters."""
    if fileMask:
      if isinstance(fileMask, list):
        fileMask = ';'.join(fileMask)
      self.setParameter('outputDataFileMask', 'string', fileMask, 'outputDataFileMask')

    if stepMask:
      if isinstance(stepMask, list):
        stepMask = ';'.join(stepMask)
      self.LHCbJob._addParameter(self.LHCbJob.workflow, 'outputDataStep', 'string', stepMask,
                                 'outputDataStep Mask')  # pylint: disable=protected-access

  #############################################################################

  def banTier1s(self):
    """Sets Tier1s as banned."""
    tier1s = []
    sites = getSites()
    if not sites['OK']:
      return sites

    for site in sites['Value']:
      tier = getSiteTier(site)
      if not tier['OK']:
        return tier
      if int(tier['Value']) in (0, 1):
        tier1s.append(site)

    self.LHCbJob.setBannedSites(tier1s)

  #############################################################################

  def banSites(self, listOfSites):
    """Sets Sites as banned."""
    sitesToBan = []
    sites = getSites()
    if not sites['OK']:
      return sites

    sites = sites['Value']
    for site in listOfSites:
      if site in sites:
        sitesToBan.append(site)

    self.LHCbJob.setBannedSites(sitesToBan)

  #############################################################################

  def setOutputMode(self, outputMode):
    """Sets output mode for all jobs, this can be 'Local' or 'Any'."""
    if not outputMode.lower().capitalize() in ('Local', 'Any', 'Run'):
      raise TypeError("Output mode must be 'Local' or 'Any' or 'Run'")
    self.setParameter('outputMode', 'string', outputMode.lower().capitalize(), 'SEResolutionPolicy')

  #############################################################################

  def setBKParameters(self, configName, configVersion, groupDescriptionOrStepsList, conditions):
    """Sets BK parameters for production."""
    self.setParameter('configName', 'string', configName, 'ConfigName')
    self.setParameter('configVersion', 'string', configVersion, 'ConfigVersion')
    if isinstance(groupDescriptionOrStepsList, list):
      # in this case we assume it is a list of steps (stepsListDict in
      # ProductionRequest module), so we calculate the BK path
      groupDescription = ''
      for step in groupDescriptionOrStepsList:
        if step['Visible'] == 'Y':
          groupDescription = os.path.sep.join([groupDescription, step['ProcessingPass']])
    else:
      groupDescription = groupDescriptionOrStepsList
    self.setParameter('groupDescription', 'string', groupDescription, 'GroupDescription')
    self.setParameter('conditions', 'string', conditions, 'SimOrDataTakingCondsString')
    self.setParameter('simDescription', 'string', conditions, 'SimDescription')

  #############################################################################
  # properties

  def set_transformationFamily(self, value):
    if isinstance(value, int):
      value = str(value)
    self._transformationFamily = value

  def get_transformationFamily(self):
    return self._transformationFamily
  startRun = property(get_transformationFamily, set_transformationFamily)
